# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from keyutil import get_obj_info_by_key


if __name__ == "__main__":
    keys = [
        "task role:geerlingguy.gitlab#taskfile:tasks/main.yml#task:[0]",
        "taskfile role:geerlingguy.gitlab#taskfile:tasks/main.yml",
        "role role:geerlingguy.gitlab",
        "play collection:debops.debops#playbook:playbooks/virt/dnsmasq-persistent_paths.yml#play:[0]",
        "playbook" " collection:debops.debops#playbook:playbooks/sys/cryptsetup-plain.yml",
        "module collection:debops.debops#module:debops.debops.apache2_module",
    ]
    for k in keys:
        info = get_obj_info_by_key(k)
        print(json.dumps(info, indent=2))



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from pathlib import Path
import re
import os
import yaml
import ansible_risk_insight.logger as logger
from .safe_glob import safe_glob
from .awx_utils import could_be_playbook, search_playbooks


module_name_re = re.compile(r"^[a-z0-9_]+\.[a-z0-9_]+\.[a-z0-9_]+$")

module_dir_patterns = [
    "library",
    "plugins/modules",
    "plugins/actions",
]

playbook_taskfile_dir_patterns = ["tasks", "playbooks"]


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass(frozen=True)
class TaskKeywordSet(metaclass=Singleton):
    task_keywords: set


@dataclass(frozen=True)
class BuiltinModuleSet(metaclass=Singleton):
    builtin_modules: set


p = Path(__file__).resolve().parent
with open(p / "task_keywords.txt", "r") as f:
    TaskKeywordSet(set(f.read().splitlines()))

with open(p / "builtin-modules.txt", "r") as f:
    BuiltinModuleSet(set(f.read().splitlines()))


def get_builtin_module_names():
    return BuiltinModuleSet().builtin_modules


def find_module_name(data_block):
    keys = [k for k in data_block.keys()]
    task_keywords = TaskKeywordSet().task_keywords
    builtin_modules = BuiltinModuleSet().builtin_modules
    for k in keys:
        if k.startswith("ansible.builtin"):
            return k
        if k in builtin_modules:
            return k
        if module_name_re.match(k):
            return k
    for k in keys:
        if k not in task_keywords and not k.startswith("with_"):
            return k
    return ""


def get_task_blocks(fpath="", yaml_str="", task_dict_list=None):
    d = None
    yaml_lines = ""
    if yaml_str:
        try:
            d = yaml.safe_load(yaml_str)
            yaml_lines = yaml_str
        except Exception as e:
            logger.debug("failed to load this yaml string to get task blocks; {}".format(e.args[0]))
            return None, None
    elif fpath:
        if not os.path.exists(fpath):
            return None, None
        with open(fpath, "r") as file:
            try:
                yaml_lines = file.read()
                d = yaml.safe_load(yaml_lines)
            except Exception as e:
                logger.debug("failed to load this yaml file to get task blocks; {}".format(e.args[0]))
                return None, None
    elif task_dict_list is not None:
        d = task_dict_list
    else:
        return None, None
    if d is None:
        return None, None
    if not isinstance(d, list):
        return None, None
    tasks = []
    for task_dict in d:
        task_dict_loop = flatten_block_tasks(task_dict)
        tasks.extend(task_dict_loop)
    return tasks, yaml_lines


# extract all tasks by flattening block tasks recursively
# a block task like below will be flattened
# like [some_module1, some_module2, some_module3]
#
# - block:
#     - some_module1:
#     - block:
#         - some_module2
#         - some_module3
#
def flatten_block_tasks(task_dict):
    if task_dict is None:
        return []
    tasks = []
    if "block" in task_dict:
        tasks_in_block = task_dict.get("block", [])
        if isinstance(tasks_in_block, list):
            for t_dict in tasks_in_block:
                tasks_in_item = flatten_block_tasks(t_dict)
                tasks.extend(tasks_in_item)
        else:
            tasks = [task_dict]
    else:
        tasks = [task_dict]
    return tasks


def search_module_files(path, module_dir_paths=[]):
    file_list = []
    # must copy the input here; otherwise, the added items are kept forever
    search_targets = [p for p in module_dir_paths]
    for module_dir_pattern in module_dir_patterns:
        search_targets.append(os.path.join(path, module_dir_pattern))
    for search_target in search_targets:
        for dirpath, folders, files in os.walk(search_target):
            for file in files:
                basename, ext = os.path.splitext(file)
                if basename == "__init__":
                    continue
                if ext == ".py" or ext == "":
                    fpath = os.path.join(dirpath, file)

                    # check if "DOCUMENTATION" is found in the file
                    skip = False
                    with open(fpath, "r") as f:
                        body = f.read()
                        if "DOCUMENTATION" not in body:
                            # if not, it is not a module file, so skip it
                            skip = True
                    if skip:
                        continue

                    file_list.append(fpath)
    file_list = sorted(file_list)
    return file_list


def find_module_dirs(role_root_dir):
    module_dirs = []
    for module_dir_pattern in module_dir_patterns:
        moddir = os.path.join(role_root_dir, module_dir_pattern)
        if os.path.exists(moddir):
            module_dirs.append(moddir)
    return module_dirs


def search_taskfiles_for_playbooks(path, taskfile_dir_paths=[]):
    # must copy the input here; otherwise, the added items are kept forever
    search_targets = [p for p in taskfile_dir_paths]
    for playbook_taskfile_dir_pattern in playbook_taskfile_dir_patterns:
        search_targets.append(os.path.join(path, playbook_taskfile_dir_pattern))
    candidates = []
    for search_target in search_targets:
        patterns = [search_target + "/**/*.yml", search_target + "/**/*.yaml"]
        found = safe_glob(patterns, recursive=True)
        for f in found:
            # taskfiles in role will be loaded when the role is loaded, so skip
            if "/roles/" in f:
                continue
            # if it is a playbook, skip it
            if could_be_playbook(f):
                continue
            d = None
            with open(f, "r") as file:
                try:
                    d = yaml.safe_load(file)
                except Exception as e:
                    logger.debug("failed to load this yaml file to search task" " files; {}".format(e.args[0]))
            # if d cannot be loaded as tasks yaml file, skip it
            if d is None or not isinstance(d, list):
                continue
            candidates.append(f)
    return candidates


def search_inventory_files(path):
    inventory_file_patterns = [
        os.path.join(path, "**/group_vars", "*"),
        os.path.join(path, "**/host_vars", "*"),
    ]
    files = safe_glob(patterns=inventory_file_patterns, recursive=True)
    return files


def find_best_repo_root_path(path):
    # get all possible playbooks
    playbooks = search_playbooks(path)
    # sort by directory depth to find the most top playbook
    playbooks = sorted(playbooks, key=lambda x: len(x.split(os.sep)))
    # still "repo/xxxxx/sample1.yml" may come before
    # "repo/playbooks/sample2.yml" because the depth are same,
    # so specifically put "playbooks" or "playbook" ones first
    if len(playbooks) > 0:
        most_shallow_depth = len(playbooks[0].split(os.sep))
        playbooks_ordered = []
        rests = []
        for p in playbooks:
            is_shortest = len(p.split(os.sep)) == most_shallow_depth
            is_playbook_dir = "/playbooks/" in p or "/playbook/" in p
            if is_shortest and is_playbook_dir:
                playbooks_ordered.append(p)
            else:
                rests.append(p)
        playbooks_ordered.extend(rests)
        playbooks = playbooks_ordered
    # ignore tests directory
    playbooks = [p for p in playbooks if "/tests/" not in p]
    if len(playbooks) == 0:
        raise ValueError("no playbook files found under {}".format(path))
    top_playbook_path = playbooks[0]
    root_path = ""
    if "/playbooks/" in top_playbook_path:
        root_path = top_playbook_path.split("/playbooks/")[0]
    elif "/playbook/" in top_playbook_path:
        root_path = top_playbook_path.split("/playbook/")[0]
    else:
        root_path = os.path.dirname(top_playbook_path)
    return root_path


def find_collection_name_of_repo(path):
    pattern = os.path.join(path, "**/galaxy.yml")
    found_galaxy_ymls = safe_glob(pattern, recursive=True)
    my_collection_name = ""
    if len(found_galaxy_ymls) > 0:
        galaxy_yml = found_galaxy_ymls[0]
        my_collection_info = None
        with open(galaxy_yml, "r") as file:
            try:
                my_collection_info = yaml.safe_load(file)
            except Exception as e:
                logger.debug("failed to load this yaml file to read galaxy.yml; {}".format(e.args[0]))
        if my_collection_info is None:
            return ""
        namespace = my_collection_info.get("namespace", "")
        name = my_collection_info.get("name", "")
        my_collection_name = "{}.{}".format(namespace, name)
    return my_collection_name



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from ansible_risk_insight.models import TaskCall, Annotation, RiskAnnotation, DefaultRiskType
from ansible_risk_insight.annotators.variable_resolver import VariableAnnotation
from ansible_risk_insight.annotators.risk_annotator_base import RiskAnnotator


class SampleCustomAnnotator(RiskAnnotator):
    name: str = "sample"
    enabled: bool = False

    # whether this task should be analyzed by this or not
    def match(self, taskcall: TaskCall) -> bool:
        # resolved_name = taskcall.resolved_name
        # return resolved_name.startswith("sample.custom.")
        return False

    # extract analyzed_data from task and embed it
    def run(self, task: TaskCall) -> List[Annotation]:
        resolved_name = task.spec.resolved_name
        options = task.spec.module_options
        var_annos = task.get_annotation_by_type(VariableAnnotation.type)
        var_anno = var_annos[0] if len(var_annos) > 0 else VariableAnnotation()
        resolved_options = var_anno.resolved_module_options

        annotations = []
        # example of package_install
        if resolved_name == "sample.custom.homebrew":
            res = RiskAnnotation(type=self.type, category=DefaultRiskType.PACKAGE_INSTALL)
            res.data = self.homebrew(options)
            for ro in resolved_options:
                res.resolved_data.append(self.homebrew(ro))
            annotations.append(res)
        return annotations

    def homebrew(self, options):
        data = {}
        if type(options) is not dict:
            return data
        if "name" in options:
            data["pkg"] = options["name"]
        if "state" in options and options["state"] == "absent":
            data["delete"] = True
        return data



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    Rule,
    Severity,
    RuleTag as Tag,
    ExecutableType,
)


@dataclass
class ModuleArgumentKeyValidationRule(Rule):
    rule_id: str = "P002"
    description: str = "Validate module argument keys and set annotations"
    enabled: bool = True
    name: str = "ModuleArgumentKeyValidation"
    version: str = "v0.0.1"
    severity: Severity = Severity.NONE
    tags: tuple = Tag.QUALITY
    precedence: int = 0

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        if task.spec.executable_type == ExecutableType.MODULE_TYPE and task.module and task.module.arguments:

            mo = task.spec.module_options
            used_keys = []
            if isinstance(mo, dict):
                used_keys = list(mo.keys())
            available_keys = []
            required_keys = []
            alias_reverse_map = {}
            available_args = None
            if task.module:
                for arg in task.module.arguments:
                    available_keys.extend(arg.available_keys())
                    if arg.required:
                        aliases = arg.aliases if arg.aliases else []
                        req_k = {"key": arg.name, "aliases": aliases}
                        required_keys.append(req_k)
                    if arg.aliases:
                        for al in arg.aliases:
                            alias_reverse_map[al] = arg.name
                available_args = task.module.arguments
            wrong_keys = [key for key in used_keys if key not in available_keys]
            missing_required_keys = []
            for k in required_keys:
                name = k.get("key", "")
                aliases = k.get("aliases", [])
                if name in used_keys:
                    continue
                if aliases:
                    found = False
                    for a_k in aliases:
                        if a_k in used_keys:
                            found = True
                            break
                    if found:
                        continue
                # here, the required key was not found
                missing_required_keys.append(name)

            used_alias_and_real_keys = []
            for k in used_keys:
                if k not in alias_reverse_map:
                    continue
                real_name = alias_reverse_map[k]
                used_alias_and_real_keys.append(
                    {
                        "used_alias": k,
                        "real_key": real_name,
                    }
                )

            task.set_annotation("module.wrong_arg_keys", wrong_keys, rule_id=self.rule_id)
            task.set_annotation("module.available_arg_keys", available_keys, rule_id=self.rule_id)
            task.set_annotation("module.required_arg_keys", required_keys, rule_id=self.rule_id)
            task.set_annotation("module.missing_required_arg_keys", missing_required_keys, rule_id=self.rule_id)
            task.set_annotation("module.available_args", available_args, rule_id=self.rule_id)
            task.set_annotation("module.used_alias_and_real_keys", used_alias_and_real_keys, rule_id=self.rule_id)

        # TODO: find duplicate keys

        return None



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import jsonpickle
from dataclasses import dataclass, field
import tarfile

from .models import (
    LoadType,
    ObjectList,
    ExecutableType,
    Module,
    ModuleMetadata,
)
from .findings import Findings
from .utils import escape_url, version_to_num, diff_files_data
from .safe_glob import safe_glob
from .keyutil import get_obj_info_by_key
from .model_loader import load_builtin_modules


module_index_name = "module_index.json"


@dataclass
class RAMClient(object):
    root_dir: str = ""

    findings_json_list_cache: list = field(default_factory=list)
    modules_json_list_cache: list = field(default_factory=list)
    roles_json_list_cache: list = field(default_factory=list)
    taskfiles_json_list_cache: list = field(default_factory=list)
    tasks_json_list_cache: list = field(default_factory=list)

    modules_cache: dict = field(default_factory=dict)
    roles_cache: dict = field(default_factory=dict)
    taskfiles_cache: dict = field(default_factory=dict)
    tasks_cache: dict = field(default_factory=dict)

    module_search_cache: dict = field(default_factory=dict)
    task_search_cache: dict = field(default_factory=dict)

    builtin_modules_cache: dict = field(default_factory=dict)

    module_index: dict = field(default_factory=dict)

    def __post_init__(self):
        module_index_path = os.path.join(self.root_dir, "indices", module_index_name)
        if os.path.exists(module_index_path):
            with open(module_index_path, "r") as file:
                self.module_index = json.load(file)

    def register(self, findings: Findings):
        metadata = findings.metadata

        type = metadata.get("type", "")
        name = metadata.get("name", "")
        version = metadata.get("version", "")
        hash = metadata.get("hash", "")

        out_dir = self.make_findings_dir_path(type, name, version, hash)
        self.save_findings(findings, out_dir)

    def module_index_register(self, findings: Findings):
        modules = self.load_module_index()
        for module in findings.root_definitions.get("definitions", {}).get("modules", []):
            if not isinstance(module, Module):
                continue
            m_meta = ModuleMetadata.from_module(module, findings.metadata)
            current = modules.get(module.name, [])
            if m_meta not in current:
                current.append(m_meta)
            modules.update({module.name: current})
        self.save_module_index(modules)

    def make_findings_dir_path(self, type, name, version, hash):
        type_root = type + "s"
        dir_name = name
        if type in [LoadType.PROJECT, LoadType.PLAYBOOK, LoadType.TASKFILE]:
            dir_name = escape_url(name)
        ver_str = version if version != "" else "unknown"
        hash_str = hash if hash != "" else "unknown"
        out_dir = os.path.join(self.root_dir, type_root, "findings", dir_name, ver_str, hash_str)
        return out_dir

    def load_metadata_from_findings(self, type, name, version, hash="*"):
        findings = self.search_findings(name, version, type)
        if not findings:
            return False, None, None
        if not isinstance(findings, Findings):
            return False, None, None
        return True, findings.metadata, findings.dependencies

    def load_definitions_from_findings(self, type, name, version, hash, allow_unresolved=False):
        findings_dir = self.make_findings_dir_path(type, name, version, hash)
        findings_path = os.path.join(findings_dir, "findings.json")
        loaded = False
        definitions = {}
        mappings = {}
        if os.path.exists(findings_path):
            findings = Findings.load(fpath=findings_path)
            # use RAM only if no unresolved dependency
            # (RAM should be fully-resolved specs as much as possible)
            if findings and (len(findings.extra_requirements) == 0 or allow_unresolved):
                definitions = findings.root_definitions.get("definitions", {})
                mappings = findings.root_definitions.get("mappings", {})
                if mappings:
                    loaded = True
        return loaded, definitions, mappings

    def search_builtin_module(self, name, used_in=""):
        builtin_modules = {}
        if self.builtin_modules_cache:
            builtin_modules = self.builtin_modules_cache
        else:
            builtin_modules = load_builtin_modules()
            self.builtin_modules_cache = builtin_modules
        short_name = name
        if "ansible.builtin." in name:
            short_name = name.split(".")[-1]
        matched_modules = []
        if short_name in builtin_modules:
            m = builtin_modules[short_name]
            matched_modules.append(
                {
                    "type": "module",
                    "name": m.fqcn,
                    "object": m,
                    "defined_in": {
                        "type": "collection",
                        "name": m.collection,
                        "version": "unknown",
                        "hash": "unknown",
                    },
                    "used_in": used_in,
                }
            )
        return matched_modules

    def load_from_indice(self, short_name, meta, used_in=""):
        _type = meta.get("type", "")
        _name = meta.get("name", "")
        collection = ""
        role = ""
        if _type == "collection":
            collection = _name
        elif _type == "role":
            role = _name
        _version = meta.get("version", "")
        _hash = meta.get("hash", "")
        m = Module(
            name=short_name,
            fqcn=meta.get("fqcn", ""),
            collection=collection,
            role=role,
        )
        m_wrapper = {
            "type": "module",
            "name": m.fqcn,
            "object": m,
            "defined_in": {
                "type": m.type,
                "name": _name,
                "version": _version,
                "hash": _hash,
            },
            "used_in": used_in,
        }
        return m_wrapper

    def search_module(self, name, exact_match=False, max_match=-1, collection_name="", collection_version="", used_in=""):
        if max_match == 0:
            return []
        args_str = json.dumps([name, exact_match, max_match, collection_name, collection_version])
        if args_str in self.module_search_cache:
            return self.module_search_cache[args_str]

        # check if the module is builtin
        matched_builtin_modules = self.search_builtin_module(name, used_in)
        if len(matched_builtin_modules) > 0:
            self.module_search_cache[args_str] = matched_builtin_modules
            return matched_builtin_modules

        short_name = name
        if "." in name:
            short_name = name.split(".")[-1]

        from_indices = False
        found_index = None
        if short_name in self.module_index and self.module_index[short_name]:
            from_indices = True
            found_index = self.module_index[short_name][0]

        modules_json_list = []
        if from_indices:
            _type = found_index.get("type", "")
            _name = found_index.get("name", "")
            _version = found_index.get("version", "")
            _hash = found_index.get("hash", "")
            findings_path = os.path.join(self.root_dir, _type + "s", "findings", _name, _version, _hash, "findings.json")
            if os.path.exists(findings_path):
                modules_json_list.append(findings_path)
        else:
            # Do not search a module from all findings
            # when it is not found in the module index.
            # Instead, just return nothing in the case.
            pass
        matched_modules = []
        search_end = False
        for findings_json in modules_json_list:
            modules = ObjectList()
            if findings_json in self.modules_cache:
                modules = self.modules_cache[findings_json]
            else:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                modules = f.root_definitions.get("definitions", {}).get("modules", [])
                self.modules_cache[findings_json] = modules
            for m in modules:
                matched = False
                if exact_match:
                    if m.fqcn == name:
                        matched = True
                else:
                    if m.fqcn == name or m.fqcn.endswith(f".{name}"):
                        matched = True
                if matched:
                    parts = findings_json.split("/")

                    matched_modules.append(
                        {
                            "type": "module",
                            "name": m.fqcn,
                            "object": m,
                            "defined_in": {
                                "type": parts[-6][:-1],  # collection or role
                                "name": parts[-4],
                                "version": parts[-3],
                                "hash": parts[-2],
                            },
                            "used_in": used_in,
                        }
                    )
                if max_match > 0:
                    if len(matched_modules) >= max_match:
                        search_end = True
                        break
            if search_end:
                break
        self.module_search_cache[args_str] = matched_modules
        return matched_modules

    def search_role(self, name, exact_match=False, max_match=-1, collection_name="", collection_version="", used_in=""):
        if max_match == 0:
            return []
        findings_json_list = []
        if self.findings_json_list_cache:
            findings_json_list = self.findings_json_list_cache
        else:
            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
            findings_json_list_coll = safe_glob(search_patterns)
            findings_json_list_coll = sort_by_version(findings_json_list_coll)
            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
            findings_json_list_role = safe_glob(search_patterns)
            findings_json_list_role = sort_by_version(findings_json_list_role)
            findings_json_list = findings_json_list_coll + findings_json_list_role
            self.findings_json_list_cache = findings_json_list

        roles_json_list = []
        if self.roles_json_list_cache:
            roles_json_list = self.roles_json_list_cache
        else:
            for findings_json in findings_json_list:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                # avoid using unresolved RAM data
                if f.extra_requirements:
                    continue
                roles = f.root_definitions.get("definitions", {}).get("roles", [])
                self.roles_cache[findings_json] = roles
                roles_json_list.append(findings_json)
            self.roles_json_list_cache = roles_json_list

        if collection_name != "":
            roles_json_list = [fpath for fpath in roles_json_list if f"/{collection_name}/" in fpath]
        if collection_version != "":
            roles_json_list = [fpath for fpath in roles_json_list if f"/{collection_version}/" in fpath]

        matched_roles = []
        search_end = False
        for findings_json in roles_json_list:
            roles = ObjectList()
            if findings_json in self.roles_cache:
                roles = self.roles_cache[findings_json]
            else:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                roles = f.root_definitions.get("definitions", {}).get("roles", [])
                self.roles_cache[findings_json] = roles
            for r in roles:
                matched = False
                if exact_match:
                    if r.fqcn == name:
                        matched = True
                else:
                    if r.fqcn == name or r.fqcn.endswith(f".{name}"):
                        matched = True
                if matched:
                    parts = findings_json.split("/")
                    offspring_objects = []
                    for taskfile_key in r.taskfiles:
                        _tmp_offspring_objects = self.search_taskfile(taskfile_key, is_key=True, collection_name=r.collection)
                        if len(_tmp_offspring_objects) > 0:
                            tf = _tmp_offspring_objects[0]
                            if tf:
                                offspring_objects.append(tf)
                            offspr_objs = _tmp_offspring_objects[0].get("offspring_objects", [])
                            if offspr_objs:
                                _offspring_obj_set = set()
                                for offspr_obj in offspr_objs:
                                    _offspr_obj_instance = offspr_obj.get("object", None)
                                    if _offspr_obj_instance is None:
                                        continue
                                    if _offspr_obj_instance.key not in _offspring_obj_set:
                                        offspring_objects.append(offspr_obj)
                                        _offspring_obj_set.add(_offspr_obj_instance.key)
                    matched_roles.append(
                        {
                            "type": "role",
                            "name": r.fqcn,
                            "object": r,
                            "offspring_objects": offspring_objects,
                            "defined_in": {
                                "type": parts[-5][:-1],  # collection or role
                                "name": parts[-4],
                                "version": parts[-3],
                                "hash": parts[-2],
                            },
                            "used_in": used_in,
                        }
                    )
                if max_match > 0:
                    if len(matched_roles) >= max_match:
                        search_end = True
                        break
            if search_end:
                break
        return matched_roles

    def search_taskfile(self, name, include_task_path="", max_match=-1, is_key=False, collection_name="", collection_version="", used_in=""):
        if max_match == 0:
            return []
        findings_json_list = []
        if self.findings_json_list_cache:
            findings_json_list = self.findings_json_list_cache
        else:
            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
            findings_json_list_coll = safe_glob(search_patterns)
            findings_json_list_coll = sort_by_version(findings_json_list_coll)
            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
            findings_json_list_role = safe_glob(search_patterns)
            findings_json_list_role = sort_by_version(findings_json_list_role)
            findings_json_list = findings_json_list_coll + findings_json_list_role
            self.findings_json_list_cache = findings_json_list

        taskfiles_json_list = []
        if self.taskfiles_json_list_cache:
            taskfiles_json_list = self.taskfiles_json_list_cache
        else:
            for findings_json in findings_json_list:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                # avoid using unresolved RAM data
                if f.extra_requirements:
                    continue
                taskfiles = f.root_definitions.get("definitions", {}).get("taskfiles", [])
                self.taskfiles_cache[findings_json] = taskfiles
                taskfiles_json_list.append(findings_json)
            self.taskfiles_json_list_cache = taskfiles_json_list

        if collection_name != "":
            taskfiles_json_list = [fpath for fpath in taskfiles_json_list if f"/{collection_name}/" in fpath]
        if collection_version != "":
            taskfiles_json_list = [fpath for fpath in taskfiles_json_list if f"/{collection_version}/" in fpath]

        search_path_list = []
        taskfile_ref = name
        if include_task_path != "":
            base_path = os.path.dirname(include_task_path)
            taskfile_path = os.path.normpath(os.path.join(base_path, taskfile_ref))
            search_path_list.append(taskfile_path)
            if "roles/" in taskfile_ref and "roles/" in base_path:
                root_path = base_path.split("roles/")[0]
                taskfile_path = os.path.normpath(os.path.join(root_path, taskfile_ref))
                search_path_list.append(taskfile_path)

        matched_taskfiles = []
        search_end = False
        for findings_json in taskfiles_json_list:
            taskfiles = ObjectList()
            if findings_json in self.taskfiles_cache:
                taskfiles = self.taskfiles_cache[findings_json]
            else:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                taskfiles = f.root_definitions.get("definitions", {}).get("taskfiles", [])
                self.taskfiles_cache[findings_json] = taskfiles
            for tf in taskfiles:
                matched = False
                if is_key:
                    if tf.key == name:
                        matched = True
                else:
                    if tf.defined_in in search_path_list:
                        matched = True
                    # TODO: support taskfile reference with variables
                if matched:
                    parts = findings_json.split("/")
                    offspring_objects = []
                    for task_key in tf.tasks:
                        _tmp_offspring_objects = self.search_task(task_key, is_key=True, collection_name=tf.collection, used_in=used_in)
                        if len(_tmp_offspring_objects) > 0:
                            t = _tmp_offspring_objects[0]
                            if t:
                                offspring_objects.append(t)
                            offspr_objs = _tmp_offspring_objects[0].get("offspring_objects", [])
                            if offspr_objs:
                                _offspring_obj_set = set()
                                for offspr_obj in offspr_objs:
                                    _offspr_obj_instance = offspr_obj.get("object", None)
                                    if _offspr_obj_instance is None:
                                        continue
                                    if _offspr_obj_instance.key not in _offspring_obj_set:
                                        offspring_objects.append(offspr_obj)
                                        _offspring_obj_set.add(_offspr_obj_instance.key)

                    matched_taskfiles.append(
                        {
                            "type": "taskfile",
                            "name": tf.key,
                            "object": tf,
                            "offspring_objects": offspring_objects,
                            "defined_in": {
                                "type": parts[-5][:-1],  # collection or role
                                "name": parts[-4],
                                "version": parts[-3],
                                "hash": parts[-2],
                            },
                            "used_in": used_in,
                        }
                    )
                if max_match > 0:
                    if len(matched_taskfiles) >= max_match:
                        search_end = True
                        break
            if search_end:
                break
        return matched_taskfiles

    def search_task(self, name, exact_match=False, max_match=-1, is_key=False, collection_name="", collection_version="", used_in=""):
        if max_match == 0:
            return []
        args_str = json.dumps([name, exact_match, max_match, is_key, collection_name, collection_version])
        if args_str in self.task_search_cache:
            return self.task_search_cache[args_str]
        findings_json_list = []
        if self.findings_json_list_cache:
            findings_json_list = self.findings_json_list_cache
        else:
            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
            findings_json_list_coll = safe_glob(search_patterns)
            findings_json_list_coll = sort_by_version(findings_json_list_coll)
            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
            findings_json_list_role = safe_glob(search_patterns)
            findings_json_list_role = sort_by_version(findings_json_list_role)
            findings_json_list = findings_json_list_coll + findings_json_list_role
            self.findings_json_list_cache = findings_json_list

        tasks_json_list = []
        if self.tasks_json_list_cache:
            tasks_json_list = self.tasks_json_list_cache
        else:
            for findings_json in findings_json_list:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                # avoid using unresolved RAM data
                if f.extra_requirements:
                    continue
                tasks = f.root_definitions.get("definitions", {}).get("tasks", [])
                self.tasks_cache[findings_json] = tasks
                tasks_json_list.append(findings_json)
            self.tasks_json_list_cache = tasks_json_list

        if collection_name != "":
            tasks_json_list = [fpath for fpath in tasks_json_list if f"/{collection_name}/" in fpath]
        if collection_version != "":
            tasks_json_list = [fpath for fpath in tasks_json_list if f"/{collection_version}/" in fpath]

        matched_tasks = []
        search_end = False
        for findings_json in tasks_json_list:
            tasks = ObjectList()
            if findings_json in self.tasks_cache:
                tasks = self.tasks_cache[findings_json]
            else:
                f = Findings.load(fpath=findings_json)
                if not isinstance(f, Findings):
                    continue
                tasks = f.root_definitions.get("definitions", {}).get("tasks", [])
                self.tasks_cache[findings_json] = tasks
            for t in tasks:
                matched = False
                if is_key:
                    if t.key == name:
                        matched = True
                else:
                    if exact_match:
                        if t.name == name:
                            matched = True
                    else:
                        if t.name == name or name in t.name:
                            matched = True
                if matched:
                    parts = findings_json.split("/")
                    offspring_objects = []
                    if t.executable_type == ExecutableType.MODULE_TYPE:
                        _tmp_offspring_objects = self.search_module(t.executable, used_in=t.defined_in)
                    elif t.executable_type == ExecutableType.ROLE_TYPE:
                        _tmp_offspring_objects = self.search_role(t.executable, used_in=t.defined_in)
                    elif t.executable_type == ExecutableType.TASKFILE_TYPE:
                        _tmp_offspring_objects = self.search_taskfile(
                            t.executable, include_task_path=t.defined_in, collection_name=t.collection, used_in=t.defined_in
                        )
                    if len(_tmp_offspring_objects) > 0:
                        child = _tmp_offspring_objects[0]
                        if child:
                            offspring_objects.append(child)
                        offspr_objs = _tmp_offspring_objects[0].get("offspring_objects", [])
                        if offspr_objs:
                            _offspring_obj_set = set()
                            for offspr_obj in offspr_objs:
                                _offspr_obj_instance = offspr_obj.get("object", None)
                                if _offspr_obj_instance is None:
                                    continue
                                if _offspr_obj_instance.key not in _offspring_obj_set:
                                    offspring_objects.append(offspr_obj)
                                    _offspring_obj_set.add(_offspr_obj_instance.key)

                    matched_tasks.append(
                        {
                            "type": "task",
                            "name": t.key,
                            "object": t,
                            "offspring_objects": offspring_objects,
                            "defined_in": {
                                "type": parts[-5][:-1],  # collection or role
                                "name": parts[-4],
                                "version": parts[-3],
                                "hash": parts[-2],
                            },
                            "used_in": used_in,
                        }
                    )
                if max_match > 0:
                    if len(matched_tasks) >= max_match:
                        search_end = True
                        break
            if search_end:
                break
        self.task_search_cache[args_str] = matched_tasks
        return matched_tasks

    def get_object_by_key(self, obj_key: str):
        obj_info = get_obj_info_by_key(obj_key)
        obj_type = obj_info.get("type", "")
        parent_name = obj_info.get("parent_name", "")
        type_str = obj_type + "s"

        search_patterns = os.path.join(self.root_dir, "collections", "findings", parent_name, "*", "*", "root", f"{type_str}.json")
        obj_json_list_coll = safe_glob(search_patterns)
        obj_json_list_coll = sort_by_version(obj_json_list_coll)
        search_patterns = os.path.join(self.root_dir, "roles", "findings", parent_name, "*", "*", "root", f"{type_str}.json")
        obj_json_list_role = safe_glob(search_patterns)
        obj_json_list_role = sort_by_version(obj_json_list_role)
        obj_json_list = obj_json_list_coll + obj_json_list_role

        matched_obj = None
        for obj_json in obj_json_list:
            objs = ObjectList.from_json(fpath=obj_json)
            obj = objs.find_by_key(obj_key)
            if obj is not None:
                parts = obj_json.split("/")
                matched_obj = {
                    "object": obj,
                    "defined_in": {
                        "type": parts[-6][:-1],  # collection or role
                        "name": parts[-5],
                        "version": parts[-4],
                        "hash": parts[-3],
                    },
                }
        return matched_obj

    def list_all_ram_metadata(self):
        findings_json_list = []
        if self.findings_json_list_cache:
            findings_json_list = self.findings_json_list_cache
        else:
            search_patterns = os.path.join(self.root_dir, "collections", "findings", "*", "*", "*", "findings.json")
            findings_json_list_coll = safe_glob(search_patterns)
            findings_json_list_coll = sort_by_version(findings_json_list_coll)
            search_patterns = os.path.join(self.root_dir, "roles", "findings", "*", "*", "*", "findings.json")
            findings_json_list_role = safe_glob(search_patterns)
            findings_json_list_role = sort_by_version(findings_json_list_role)
            findings_json_list = findings_json_list_coll + findings_json_list_role
            self.findings_json_list_cache = findings_json_list

        metadata_list = []
        for findings_path in findings_json_list:
            parts = findings_path.split("/")

            metadata_list.append(
                {
                    "type": parts[-5][:-1],  # collection or role
                    "name": parts[-4],
                    "version": parts[-3],
                    "hash": parts[-2],
                }
            )
        return metadata_list

    def search_findings(self, target_name, target_version, target_type=None):
        if not target_name:
            raise ValueError("target name must be specified for searching RAM data")
        if not target_version:
            target_version = "*"
        search_patterns = []
        if not target_type or target_type == "collection":
            search_patterns.append(os.path.join(self.root_dir, "collections", "findings", target_name, target_version, "*", "findings.json"))
        if not target_type or target_type == "role":
            search_patterns.append(os.path.join(self.root_dir, "roles", "findings", target_name, target_version, "*", "findings.json"))
        if not target_type or target_type == "project":
            e_target_name = escape_url(target_name)
            search_patterns.append(os.path.join(self.root_dir, "projects", "findings", e_target_name, target_version, "*", "findings.json"))
        if not target_type or target_type == "playbook":
            e_target_name = escape_url(target_name)
            search_patterns.append(os.path.join(self.root_dir, "playbooks", "findings", e_target_name, target_version, "*", "findings.json"))
        if not target_type or target_type == "taskfile":
            e_target_name = escape_url(target_name)
            search_patterns.append(os.path.join(self.root_dir, "taskfiles", "findings", e_target_name, target_version, "*", "findings.json"))
        found_path_list = safe_glob(search_patterns)
        latest_findings_path = ""
        if len(found_path_list) == 1:
            latest_findings_path = found_path_list[0]
        elif len(found_path_list) > 1:
            latest_findings_path = found_path_list[0]
            mtime = os.path.getmtime(latest_findings_path)
            for fpath in found_path_list:
                tmp_mtime = os.path.getmtime(fpath)
                if tmp_mtime > mtime:
                    latest_findings_path = fpath
                    mtime = tmp_mtime
        findings = None
        if os.path.exists(latest_findings_path):
            findings = self.load_findings(latest_findings_path)
        return findings

    def load_findings(self, path: str):

        basename = os.path.basename(path)
        dir_path = path
        if basename == "findings.json":
            dir_path = os.path.dirname(path)

        findings = Findings.load(fpath=os.path.join(dir_path, "findings.json"))
        return findings

    def save_findings(self, findings: Findings, out_dir: str):
        if out_dir == "":
            raise ValueError("output dir must be a non-empty value")

        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        findings.dump(fpath=os.path.join(out_dir, "findings.json"))

    def save_module_index(self, modules):
        out_dir = os.path.join(self.root_dir, "indices")
        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)
        modules_str = jsonpickle.encode(modules, make_refs=False, unpicklable=False)
        with open(os.path.join(out_dir, "module_index.json"), "w") as file:
            file.write(modules_str)

    def load_module_index(self):
        path = os.path.join(self.root_dir, "indices", "module_index.json")
        modules = {}
        if os.path.exists(path):
            with open(path, "r") as file:
                modules = json.load(file)
        return modules

    def save_error(self, error: str, out_dir: str):
        if out_dir == "":
            raise ValueError("output dir must be a non-empty value")

        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        with open(os.path.join(out_dir, "error.log"), "w") as file:
            file.write(error)

    def diff(self, target_name, version1, version2):
        findings1 = self.search_findings(target_name=target_name, target_version=version1)
        if not findings1:
            raise ValueError(f"{target_name}:{version1} is not found in RAM")

        findings2 = self.search_findings(target_name=target_name, target_version=version2)
        if not findings2:
            raise ValueError(f"{target_name}:{version2} is not found in RAM")

        coll_defs1 = findings1.root_definitions.get("definitions", {}).get("collections", [])
        coll_defs2 = findings2.root_definitions.get("definitions", {}).get("collections", [])

        files1 = None
        files2 = None
        if len(coll_defs1) > 0:
            files1 = coll_defs1[0].files
        if len(coll_defs2) > 0:
            files2 = coll_defs2[0].files

        if not files1:
            raise ValueError(f"Files data of {target_name}:{version1} is not recorded")

        if not files2:
            raise ValueError(f"Files data of {target_name}:{version2} is not recorded")

        return diff_files_data(files1, files2)

    def release(self, outfile):
        indices = os.path.join(self.root_dir, "indices")
        collection_findings = os.path.join(self.root_dir, "collections", "findings")
        role_findings = os.path.join(self.root_dir, "roles", "findings")
        with tarfile.open(outfile, 'w:gz') as tar:
            if os.path.exists(indices):
                tar.add(indices, arcname="indices")
            if os.path.exists(collection_findings):
                tar.add(collection_findings, arcname="collections/findings")
            if os.path.exists(role_findings):
                tar.add(role_findings, arcname="roles/findings")


# newer version comes earlier, so version num should be sorted in a reversed order
def _path_to_reversed_version_num(path):
    version = path.split("/findings/")[-1].split("/")[1]
    return -1 * version_to_num(version)


def _path_to_collection_name(path):
    collection = path.split("/findings/")[-1].split("/")[0]
    return collection


# the latest known version comes first
# `unknown` is the last
def sort_by_version(path_list):
    return sorted(path_list, key=lambda x: (_path_to_collection_name(x), _path_to_reversed_version_num(x)))



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    Rule,
    Severity,
    RuleTag as Tag,
    ExecutableType,
)


@dataclass
class ModuleNameValidationRule(Rule):
    rule_id: str = "P001"
    description: str = "Validate a module name and set annotations"
    enabled: bool = True
    name: str = "ModuleNameValidation"
    version: str = "v0.0.1"
    severity: Severity = Severity.NONE
    tags: tuple = Tag.QUALITY
    precedence: int = 0

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        suggested_fqcns = []
        suggested_dependency = []
        resolved_fqcn = ""
        wrong_module_name = ""
        not_exist = False
        correct_fqcn = ""
        need_correction = False

        # normal task
        if task.spec.executable_type == ExecutableType.MODULE_TYPE:

            resolved_fqcn = task.spec.resolved_name
            suggested_fqcns = [cand[0] for cand in task.spec.possible_candidates]
            suggested_dependency = [cand[1] for cand in task.spec.possible_candidates]

            if not task.spec.resolved_name:
                for suggestion in suggested_fqcns:
                    if not suggestion.endswith(f".{task.spec.module}"):
                        wrong_module_name = task.spec.module
                        break
                if not task.spec.possible_candidates:
                    not_exist = True
                    wrong_module_name = task.spec.module

            if task.spec.resolved_name:
                correct_fqcn = task.spec.resolved_name
            elif suggested_fqcns:
                correct_fqcn = suggested_fqcns[0]

            if correct_fqcn != task.spec.module or not_exist:
                need_correction = True

        # include_role, import_role
        elif task.spec.executable_type == ExecutableType.ROLE_TYPE:
            if "ansible.builtin." not in task.spec.module:
                resolved_fqcn = "ansible.builtin." + task.spec.module
                correct_fqcn = resolved_fqcn
                need_correction = True

        # include_tasks, import_tasks
        elif task.spec.executable_type == ExecutableType.TASKFILE_TYPE:
            if "ansible.builtin." not in task.spec.module:
                resolved_fqcn = "ansible.builtin." + task.spec.module
                correct_fqcn = resolved_fqcn
                need_correction = True

        module_examples = ""
        if task.module:
            module_examples = task.module.examples

        task.set_annotation("module.suggested_fqcn", suggested_fqcns, rule_id=self.rule_id)
        task.set_annotation("module.suggested_dependency", suggested_dependency, rule_id=self.rule_id)
        task.set_annotation("module.resolved_fqcn", resolved_fqcn, rule_id=self.rule_id)
        task.set_annotation("module.wrong_module_name", wrong_module_name, rule_id=self.rule_id)
        task.set_annotation("module.not_exist", not_exist, rule_id=self.rule_id)
        task.set_annotation("module.correct_fqcn", correct_fqcn, rule_id=self.rule_id)
        task.set_annotation("module.need_correction", need_correction, rule_id=self.rule_id)

        task.set_annotation("module.examples", module_examples, rule_id=self.rule_id)

        return None



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from ...ram_generator import RiskAssessmentModelGenerator as RAMGenerator


class RAMUpdateCLI:
    args = None

    def __init__(self):
        parser = argparse.ArgumentParser(description="TODO")
        parser.add_argument("target_type", help="content type", choices={"ram"})
        parser.add_argument("action", help="action for RAM command or target_name of search action")
        parser.add_argument("-f", "--file", help='target list like "collection community.general"')
        parser.add_argument("-r", "--resume", help="line number to resume scanning")
        args = parser.parse_args()
        self.args = args

    def run(self):
        args = self.args
        action = args.action
        if action != "update":
            raise ValueError('RAMUpdateCLI cannot be executed without "update" action')

        target_list = []
        with open(args.file, "r") as file:
            for line in file:
                parts = line.replace("\n", "").split(" ")
                if len(parts) != 2:
                    raise ValueError('target list file must be lines of "<type> <name>" such as "collection community.general"')
                target_list.append((parts[0], parts[1]))

        resume = -1
        if args.resume:
            resume = int(args.resume)
        ram_generator = RAMGenerator(target_list, resume, update=True)
        ram_generator.run()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import traceback
import subprocess
import requests
import hashlib
import yaml
import json
from copy import deepcopy
from tabulate import tabulate
from inspect import isclass
from importlib.util import spec_from_file_location, module_from_spec

import ansible_risk_insight.logger as logger
from .findings import Findings


def install_galaxy_target(target, target_type, output_dir, source_repository="", target_version=""):
    server_option = ""
    if source_repository:
        server_option = "--server {}".format(source_repository)
    target_name = target
    if target_version:
        target_name = f"{target}:{target_version}"
    logger.debug("exec ansible-galaxy cmd: ansible-galaxy {} install {} {} -p {}".format(target_type, target_name, server_option, output_dir))
    proc = subprocess.run(
        "ansible-galaxy {} install {} {} -p {}".format(target_type, target_name, server_option, output_dir),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    return proc.stdout


def install_github_target(target, output_dir):
    proc = subprocess.run(
        "git clone {} {}".format(target, output_dir),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    return proc.stdout


def get_download_metadata(typ: str, install_msg: str):
    download_url = ""
    version = ""
    if typ == "collection":
        for line in install_msg.splitlines():
            if line.startswith("Downloading "):
                download_url = line.split(" ")[1]
                version = download_url.split("-")[-1].replace(".tar.gz", "")
                break
    elif typ == "role":
        for line in install_msg.splitlines():
            if line.startswith("- downloading role from "):
                download_url = line.split(" ")[-1]
                version = download_url.split("/")[-1].replace(".tar.gz", "")
                break
    hash = ""
    if download_url != "":
        hash = get_hash_of_url(download_url)
    return download_url, version, hash


def get_installed_metadata(type, name, path, dep_dir=None):
    if dep_dir:
        dep_dir_alt = os.path.join(dep_dir, "ansible_collections")
        if os.path.exists(dep_dir_alt):
            dep_dir = dep_dir_alt
        parts = name.split(".")
        if len(parts) == 1:
            parts.append("dummy")
        dep_dir_target_path = os.path.join(dep_dir, parts[0], parts[1])
        download_url, version = get_installed_metadata(type, name, dep_dir_target_path)
        if download_url or version:
            return download_url, version
    download_url = ""
    version = ""
    galaxy_yml = "GALAXY.yml"
    galaxy_data = None
    if type == "collection":
        base_dir = "/".join(path.split("/")[:-2])
        dirs = os.listdir(base_dir)
        for dir_name in dirs:
            tmp_galaxy_data = None
            if dir_name.startswith(name) and dir_name.endswith(".info"):
                galaxy_yml_path = os.path.join(base_dir, dir_name, galaxy_yml)
                try:
                    with open(galaxy_yml_path, "r") as galaxy_yml_file:
                        tmp_galaxy_data = yaml.safe_load(galaxy_yml_file)
                except Exception:
                    pass
            if isinstance(tmp_galaxy_data, dict):
                galaxy_data = tmp_galaxy_data
    if galaxy_data is not None:
        download_url = galaxy_data.get("download_url", "")
        version = galaxy_data.get("version", "")
    return download_url, version


def get_collection_metadata(path: str):
    if not os.path.exists(path):
        return None
    manifest_json_path = os.path.join(path, "MANIFEST.json")
    meta = None
    if os.path.exists(manifest_json_path):
        with open(manifest_json_path, "r") as file:
            meta = json.load(file)
    return meta


def get_role_metadata(path: str):
    if not os.path.exists(path):
        return None
    meta_main_yml_path = os.path.join(path, "meta", "main.yml")
    meta = None
    if os.path.exists(meta_main_yml_path):
        with open(meta_main_yml_path, "r") as file:
            meta = yaml.safe_load(file)
    return meta


def escape_url(url: str):
    base_url = url.split("?")[0]
    replaced = base_url.replace("://", "__").replace("/", "_")
    return replaced


def escape_local_path(path: str):
    replaced = path.replace("/", "__")
    return replaced


def get_hash_of_url(url: str):
    response = requests.get(url)
    hash = hashlib.sha256(response.content).hexdigest()
    return hash


def split_name_and_version(target_name):
    name = target_name
    version = ""
    if ":" in target_name:
        parts = target_name.split(":")
        name = parts[0]
        version = parts[1]
    return name, version


def split_target_playbook_fullpath(fullpath: str):
    basedir = os.path.dirname(fullpath)
    if "/playbooks/" in fullpath:
        basedir = fullpath.split("/playbooks/")[0]
    target_playbook_path = fullpath.replace(basedir, "")
    if target_playbook_path[0] == "/":
        target_playbook_path = target_playbook_path[1:]
    return basedir, target_playbook_path


def split_target_taskfile_fullpath(fullpath: str):
    basedir = os.path.dirname(fullpath)
    if "/roles/" in fullpath:
        basedir = fullpath.split("/roles/")[0]
    target_taskfile_path = fullpath.replace(basedir, "")
    if target_taskfile_path[0] == "/":
        target_taskfile_path = target_taskfile_path[1:]
    return basedir, target_taskfile_path


def version_to_num(ver: str):
    if ver == "unknown":
        return 0
    # version string can be 1.2.3-abcdxyz
    ver_num_part = ver.split("-")[0]
    parts = ver_num_part.split(".")
    num = 0
    if len(parts) >= 1:
        if parts[0].isnumeric():
            num += float(parts[0])
    if len(parts) >= 2:
        if parts[1].isnumeric():
            num += float(parts[1]) * (0.001**1)
    if len(parts) >= 3:
        if parts[2].isnumeric():
            num += float(parts[2]) * (0.001**2)
    return num


def is_url(txt: str):
    return "://" in txt


def is_local_path(txt: str):
    if is_url(txt):
        return False
    if "/" in txt:
        return True
    if os.path.exists(txt):
        return True


def indent(multi_line_txt, level=0):
    lines = multi_line_txt.splitlines()
    lines = [" " * level + line for line in lines if line.replace(" ", "") != ""]
    return "\n".join(lines)


def report_to_display(data_report: dict):
    playbook_num_total = data_report["summary"].get("playbooks", {}).get("total", 0)
    # playbook_num_risk_found = data_report["summary"].get("playbooks", {}).get("risk_found", 0)
    role_num_total = data_report["summary"].get("roles", {}).get("total", 0)
    # role_num_risk_found = data_report["summary"].get("roles", {}).get("risk_found", 0)

    output_txt = ""
    output_txt += "-" * 90 + "\n"
    output_txt += "Ansible Risk Insight Report\n"
    output_txt += "-" * 90 + "\n"

    if playbook_num_total + role_num_total == 0:
        output_txt += "No playbooks and roles found\n"
    else:
        found_contents = ""
        if playbook_num_total > 0:
            found_contents += f"{playbook_num_total} playbooks"

        if role_num_total > 0:
            if found_contents != "":
                found_contents += " and "
            found_contents += f"{role_num_total} roles"

        output_txt += f"{found_contents} found\n"

    output_txt += "-" * 90 + "\n"

    report_num = 1
    for detail in data_report["details"]:
        output_txt_for_this_tree = ""
        do_report = False
        # output_txt_for_this_tree += "#{} {} - {}\n".format(report_num, tree_root_type.upper(), tree_root_name)
        results_list = detail.get("results", [])

        for result_info in results_list:
            output = result_info.get("output", "")
            if output == "":
                continue
            do_report = True
            # output_txt_for_this_tree += rule_name + "\n"
            output_txt_for_this_tree += indent(output, 0) + "\n"
        output_txt_for_this_tree += "-" * 90 + "\n"
        if do_report:
            output_txt += output_txt_for_this_tree
            report_num += 1
    return output_txt


def summarize_findings(findings: Findings, show_all: bool = False):
    metadata = findings.metadata
    dependencies = findings.dependencies
    report = findings.report
    resolve_failures = findings.resolve_failures
    extra_requirements = findings.extra_requirements
    return summarize_findings_data(metadata, dependencies, report, resolve_failures, extra_requirements, show_all)


def summarize_findings_data(metadata, dependencies, report, resolve_failures, extra_requirements, show_all: bool = False):
    target_name = metadata.get("name", "")
    output_lines = []

    report_txt = report_to_display(report)
    output_lines.append(report_txt)

    if len(dependencies) > 0:
        output_lines.append("External Dependencies")
        dep_table = [("NAME", "VERSION", "HASH")]
        for dep_info in dependencies:
            dep_meta = dep_info.get("metadata", {})
            dep_name = dep_meta.get("name", "")
            if dep_name == target_name:
                continue
            dep_version = dep_meta.get("version", "")
            dep_hash = dep_meta.get("hash", "")
            dep_table.append((dep_name, dep_version, dep_hash))
        output_lines.append(tabulate(dep_table))

    #     print("-" * 90)
    #     print("ARI scan completed!")
    #     print(f"Findings have been saved at: {self.ram_client.make_findings_dir_path(self.type, self.name, self.version, self.hash)}")
    #     print("-" * 90)

    module_failures = resolve_failures.get("module", {})
    role_failures = resolve_failures.get("role", {})
    taskfile_failures = resolve_failures.get("taskfile", {})
    module_fail_num = len(module_failures)
    role_fail_num = len(role_failures)
    taskfile_fail_num = len(taskfile_failures)
    total_fail_num = module_fail_num + role_fail_num + taskfile_fail_num
    if total_fail_num > 0:
        output_lines.append(f"Failed to resolve {module_fail_num} modules, {role_fail_num} roles, {taskfile_fail_num} taskfiles")
    if module_fail_num > 0:
        output_lines.append("- modules: ")
        for module_action in module_failures:
            called_num = module_failures[module_action]
            output_lines.append(f"  - {module_action}    ({called_num} times called)")
    if role_fail_num > 0:
        output_lines.append("- roles: ")
        for role_action in role_failures:
            called_num = role_failures[role_action]
            output_lines.append(f"  - {role_action}    ({called_num} times called)")
    if taskfile_fail_num > 0:
        output_lines.append("- taskfiles: ")
        for taskfile_action in taskfile_failures:
            called_num = taskfile_failures[taskfile_action]
            output_lines.append(f"  - {taskfile_action}    ({called_num} times called)")

    # roles = set()
    if len(extra_requirements) > 0:
        unresolved_modules = []
        unresolved_roles = []
        suggestion = {}
        for ext_req in extra_requirements:
            if ext_req.get("type", "") not in ["role", "module"]:
                continue
            req_name = ext_req.get("defined_in", {}).get("name", None)
            if req_name is None:
                continue
            if req_name == target_name:
                continue
            # print(f"[DEBUG] requirement: {ext_req}")

            obj_type = ext_req.get("type", "")
            obj_name = ext_req.get("name", "")
            short_name = obj_name.replace(f"{req_name}.", "")

            if obj_type == "module":
                unresolved_modules.append(ext_req)
            if obj_type == "role":
                unresolved_roles.append(ext_req)

            req_version = ext_req.get("defined_in", {}).get("version", None)
            req_str = json.dumps([req_name, req_version])

            if req_str not in suggestion:
                suggestion[req_str] = {"module": [], "role": []}
            suggestion[req_str][obj_type].append(short_name)

        if len(unresolved_modules) > 0:
            output_lines.append("Unresolved modules:")
            table = [("NAME", "USED_IN")]
            thresh = 4
            for ext_req in unresolved_modules[:thresh]:
                obj_name = ext_req.get("name", "")
                used_in = ext_req.get("used_in", "")
                req_name = ext_req.get("defined_in", {}).get("name", None)
                short_name = obj_name.replace(f"{req_name}.", "")
                table.append((short_name, used_in))
            if len(unresolved_modules) > thresh:
                rest_num = len(unresolved_modules) - thresh
                table.append(("", f"... and {rest_num} other modules"))
            output_lines.append(tabulate(table))

        if len(unresolved_roles) > 0:
            output_lines.append("Unresolved roles:")
            table = [("NAME", "USED_IN")]
            thresh = 4
            for ext_req in unresolved_roles[:thresh]:
                obj_name = ext_req.get("name", "")
                used_in = ext_req.get("used_in", "")
                req_name = ext_req.get("defined_in", {}).get("name", None)
                short_name = obj_name.replace(f"{req_name}.", "")
                table.append((short_name, used_in))
            if len(unresolved_roles) > thresh:
                rest_num = len(unresolved_roles) - thresh
                table.append(("", f"... and {rest_num} other roles"))
            output_lines.append(tabulate(table))

        req_name_keys = sorted(list(suggestion.keys()))
        output_lines.append("")
        output_lines.append("-- Suggested Dependencies --")
        table_data = [("NAME", "VERSION", "SUGGESTED_FOR")]
        for req_str in req_name_keys:
            req_dict = suggestion[req_str]
            req_module_list = req_dict["module"]
            req_module_num = len(req_module_list)

            req_role_list = req_dict["role"]
            req_role_num = len(req_role_list)
            if req_module_num + req_role_num == 0:
                continue

            req_info = json.loads(req_str)
            req_name = req_info[0]
            req_version = req_info[1]

            summary_str = ""
            thresh = 3
            if req_module_num > 0:
                module_names = ", ".join(req_module_list[:thresh])
                if req_module_num > thresh:
                    module_names += ", etc."
                prefix = "module" if req_module_num == 1 else "modules"
                module_names += f" (total {req_module_num} {prefix})"
                summary_str += module_names
            if req_role_num > 0:
                if summary_str != "":
                    summary_str += " and "

                role_names = ", ".join(req_role_list[:thresh])
                if req_role_num > thresh:
                    role_names += ", etc."
                prefix = "role" if req_role_num == 1 else "roles"
                role_names += f" (total {req_role_num} {prefix})"
                summary_str += role_names
            table_data.append((req_name, req_version, summary_str))
        output_lines.append(tabulate(table_data))
    output = "\n".join(output_lines)
    return output


def show_all_ram_metadata(ram_meta_list):
    table = [("NAME", "VERSION", "HASH")]
    for meta in ram_meta_list:
        table.append((meta["name"], meta["version"], meta["hash"]))
    print(tabulate(table))


def diff_files_data(files1, files2):
    files_dict1 = {}
    for finfo in files1.get("files", []):
        ftype = finfo.get("ftype", "")
        if ftype != "file":
            continue
        fpath = finfo.get("name", "")
        hash = finfo.get("chksum_sha256", "")
        files_dict1[fpath] = hash

    files_dict2 = {}
    for finfo in files2.get("files", []):
        ftype = finfo.get("ftype", "")
        if ftype != "file":
            continue
        fpath = finfo.get("name", "")
        hash = finfo.get("chksum_sha256", "")
        files_dict2[fpath] = hash

    # TODO: support "replaced" type
    diffs = []
    for fpath, hash in files_dict1.items():
        if fpath in files_dict2:
            if files_dict2[fpath] == hash:
                continue
            else:
                diffs.append(
                    {
                        "type": "updated",
                        "filepath": fpath,
                    }
                )
        else:
            diffs.append(
                {
                    "type": "created",
                    "filepath": fpath,
                }
            )

    for fpath, hash in files_dict2.items():
        if fpath in files_dict1:
            continue
        else:
            diffs.append(
                {
                    "type": "deleted",
                    "filepath": fpath,
                }
            )

    return diffs


def show_diffs(diffs):
    table = [("NAME", "DIFF_TYPE")]
    for d in diffs:
        table.append((d["filepath"], d["type"]))
    print(tabulate(table))


def get_module_specs_by_ansible_doc(module_files: str, fqcn_prefix: str, search_path: str):
    if not module_files:
        return {}

    if search_path and fqcn_prefix:
        parent_path_pattern = "/" + fqcn_prefix.replace(".", "/")
        if parent_path_pattern in search_path:
            search_path = search_path.split(parent_path_pattern)[0]

    fqcn_list = []
    for module_file_path in module_files:
        module_name = os.path.basename(module_file_path)
        if module_name[-3:] == ".py":
            module_name = module_name[:-3]
        if module_name == "__init__":
            continue
        fqcn = module_name
        if fqcn_prefix:
            fqcn = fqcn_prefix + "." + module_name
        fqcn_list.append(fqcn)
    if not fqcn_list:
        return {}
    fqcn_list_str = " ".join(fqcn_list)
    cmd_args = [f"ansible-doc {fqcn_list_str} --json"]
    _env = os.environ.copy()
    _env["ANSIBLE_COLLECTIONS_PATH"] = search_path
    proc = subprocess.run(cmd_args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, env=_env)
    if proc.stderr and not proc.stdout:
        logger.debug(f"error while getting the documentation for modules `{fqcn_list_str}`: {proc.stderr}")
        return ""
    wrapper_dict = json.loads(proc.stdout)
    specs = {}
    for fqcn in wrapper_dict:
        doc_dict = wrapper_dict[fqcn].get("doc", {})
        doc = yaml.safe_dump(doc_dict)
        examples = wrapper_dict[fqcn].get("examples", "")
        specs[fqcn] = {
            "doc": doc,
            "examples": examples,
        }
    return specs


def get_documentation_in_module_file(fpath: str):
    if not fpath:
        return ""
    if not os.path.exists(fpath):
        return ""
    lines = []
    with open(fpath, "r") as file:
        for line in file:
            lines.append(line)
    doc_lines = []
    is_inside_doc = False
    quotation = ""
    for line in lines:
        stripped_line = line.strip()

        if is_inside_doc and quotation and stripped_line.startswith(quotation):
            is_inside_doc = False
            break

        if is_inside_doc:
            if quotation:
                doc_lines.append(line)
            else:
                if "'''" in line:
                    quotation = "'''"
                if '"""' in line:
                    quotation = '"""'

        if stripped_line.startswith("DOCUMENTATION"):
            is_inside_doc = True
            if "'''" in line:
                quotation = "'''"
            if '"""' in line:
                quotation = '"""'
    return "\n".join(doc_lines)


def get_class_by_arg_type(arg_type: str):
    if not isinstance(arg_type, str):
        return None

    mapping = {
        "str": str,
        "list": list,
        "dict": dict,
        "bool": bool,
        "int": int,
        "float": float,
        # ARI handles `path` as a string
        "path": str,
        "raw": any,
        # TODO: check actual types of the following
        "jsonarg": str,
        "json": str,
        "bytes": str,
        "bits": str,
    }

    if arg_type not in mapping:
        return None

    return mapping[arg_type]


def load_classes_in_dir(dir_path: str, target_class: type, base_dir: str = "", only_subclass: bool = True, fail_on_error: bool = False):
    search_path = dir_path
    found = False
    if os.path.exists(search_path):
        found = True
    if not found and base_dir:
        self_path = os.path.abspath(base_dir)
        search_path = os.path.join(os.path.dirname(self_path), dir_path)
        if os.path.exists(search_path):
            found = True

    if not found:
        raise ValueError(f'Path not found "{dir_path}"')

    files = os.listdir(search_path)
    scripts = [os.path.join(search_path, f) for f in files if f[-3:] == ".py"]
    classes = []
    errors = []
    for s in scripts:
        try:
            short_module_name = os.path.basename(s)[:-3]
            spec = spec_from_file_location(short_module_name, s)
            mod = module_from_spec(spec)
            spec.loader.exec_module(mod)
            for k in mod.__dict__:
                cls = getattr(mod, k)
                if not callable(cls):
                    continue
                if not isclass(cls):
                    continue
                if not issubclass(cls, target_class):
                    continue
                if only_subclass and cls == target_class:
                    continue
                classes.append(cls)
        except Exception:
            exc = traceback.format_exc()
            msg = f"failed to load a rule module {s}: {exc}"
            if fail_on_error:
                raise ValueError(msg)
            else:
                errors.append(msg)
    return classes, errors


def equal(a: any, b: any):
    type_a = type(a)
    type_b = type(b)
    if type_a != type_b:
        return False
    if type_a == dict:
        all_keys = list(a.keys()) + list(b.keys())
        for key in all_keys:
            val_a = a.get(key, None)
            val_b = b.get(key, None)
            if not equal(val_a, val_b):
                return False
    elif type_a == list:
        if len(a) != len(b):
            return False
        for i in range(len(a)):
            val_a = a[i]
            val_b = b[i]
            if not equal(val_a, val_b):
                return False
    elif hasattr(a, "__dict__"):
        if not equal(a.__dict__, b.__dict__):
            return False
    else:
        if a != b:
            return False
    return True


def recursive_copy_dict(src, dst):
    if not isinstance(src, dict):
        raise ValueError(f"only dict input is allowed, but got {type(src)}")

    if not isinstance(dst, dict):
        raise ValueError(f"only dict input is allowed, but got {type(dst)}")

    for k, sv in src.items():
        if isinstance(sv, dict):
            if k not in dst:
                dst[k] = {}
            recursive_copy_dict(sv, dst[k])
        else:
            dst[k] = deepcopy(sv)
    return



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
import joblib
import os
import json
import datetime

from .scanner import ARIScanner, config


class RiskAssessmentModelGenerator(object):
    _queue: list = []
    _resume: int = -1
    _update: bool = False

    def __init__(self, target_list=[], resume=-1, update=False, parallel=True, download_only=False, no_module_spec=False, no_retry=False):
        self._queue = target_list
        self._resume = resume
        self._update = update
        self._parallel = parallel
        self._download_only = download_only
        self._no_module_spec = no_module_spec
        self._no_retry = no_retry

        use_ansible_doc = True
        if self._no_module_spec:
            use_ansible_doc = False

        self._scanner = ARIScanner(
            root_dir=config.data_dir,
            silent=True,
            use_ansible_doc=use_ansible_doc,
            read_ram=False,
        )

    def run(self):
        num = len(self._queue)
        resume_str = f"(resume from {self._resume})" if self._resume > 0 else ""
        print(f"Start scanning {num} targets {resume_str}")

        input_list = []
        for i, target_info in enumerate(self._queue):
            if i + 1 < self._resume:
                continue
            if not isinstance(target_info, tuple):
                raise ValueError(f"target list must be a list of tuple(target_type, target_name), but got a {type(target_info)}")
            if len(target_info) != 2:
                raise ValueError(f"target list must be a list of tuple(target_type, target_name), but got this; {target_info}")

            _type, _name = target_info
            input_list.append((i, num, _type, _name))

        if self._parallel:
            joblib.Parallel(n_jobs=-1)(joblib.delayed(self.scan)(i, num, _type, _name) for (i, num, _type, _name) in input_list)
        else:
            for (i, num, _type, _name) in input_list:
                self.scan(i, num, _type, _name)

    def scan(self, i, num, type, name):
        print(f"[{i+1}/{num}] {type} {name}")
        use_src_cache = True

        if self.skip_scan(type, name):
            print(f"skip ram update of {type} {name}")
            return
        fail = False
        if self._update:
            # disable dependency cache when update mode to avoid using the old src
            use_src_cache = False
        try:

            self._scanner.evaluate(
                type=type,
                name=name,
                install_dependencies=True,
                download_only=self._download_only,
                use_src_cache=use_src_cache,
            )
        except Exception:
            error = traceback.format_exc()
            self._scanner.save_error(error)
            fail = True
        self.save_ram_log(type, name, fail)

    def save_ram_log(self, type, name, fail):
        out_dir = os.path.join(self._scanner.root_dir, "log")
        path = os.path.join(out_dir, "ram_log.json")

        scan_time = datetime.datetime.utcnow().isoformat()
        new_record = {"type": type, "name": name, "succeed": not fail, "time": scan_time}

        ram_logs = {"collection": {}, "role": {}}
        if os.path.exists(path):
            with open(path, "r") as file:
                ram_logs = json.load(file)
            log = ram_logs.get(type, {}).get(name, [])
            log.append(new_record)
            ram_logs[type].update({name: log})
        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        with open(path, "w") as file:
            json.dump(ram_logs, file)
        return

    def skip_scan(self, type, name) -> bool:
        skip = False
        path = os.path.join(self._scanner.root_dir, "log", "ram_log.json")
        if not os.path.exists(path):
            return skip
        with open(path, "r") as file:
            ram_logs = json.load(file)
        log = ram_logs.get(type, {}).get(name, [])
        if log:
            latest = log[-1]
            if latest.get("succeed", False):
                skip = True
                return skip
            elif self._no_retry:
                skip = True
                return skip
        return skip



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
import jsonpickle


@dataclass
class Findings:
    metadata: dict = field(default_factory=dict)
    dependencies: list = field(default_factory=list)

    root_definitions: dict = field(default_factory=dict)
    ext_definitions: dict = field(default_factory=dict)
    extra_requirements: list = field(default_factory=list)
    resolve_failures: dict = field(default_factory=dict)

    prm: dict = field(default_factory=dict)
    report: dict = field(default_factory=dict)

    summary_txt: str = ""
    scan_time: str = ""

    def simple(self):
        d = self.report.copy()
        d["metadata"] = self.metadata
        d["dependencies"] = self.dependencies
        return d

    def dump(self, fpath=""):
        json_str = jsonpickle.encode(self, make_refs=False)
        if fpath:
            with open(fpath, "w") as file:
                file.write(json_str)
        return json_str

    def save_rule_result(self, fpath=""):
        json_str = jsonpickle.encode(self.report.get("ari_result", {}), make_refs=False, unpicklable=False)
        if fpath:
            with open(fpath, "w") as file:
                file.write(json_str)
        return json_str

    @staticmethod
    def load(fpath="", json_str=""):
        if fpath:
            with open(fpath, "r") as file:
                json_str = file.read()
        findings = jsonpickle.decode(json_str)
        return findings



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from ...scanner import config
from ...risk_assessment_model import RAMClient
from ...utils import split_name_and_version


class RAMSearchCLI:
    args = None

    def __init__(self):
        parser = argparse.ArgumentParser(description="TODO")
        parser.add_argument("target_type", help="content type", choices={"ram"})
        parser.add_argument("action", help="action for RAM command or target_name of search action")
        parser.add_argument("target_name", help="target_name for the action")
        args = parser.parse_args()
        self.args = args

    def run(self):
        args = self.args
        action = args.action
        target_name = args.target_name
        if action != "search":
            raise ValueError('RAMSearchCLI cannot be executed without "search" action')

        ram_client = RAMClient(root_dir=config.data_dir)

        target_name, target_version = split_name_and_version(target_name)
        findings = ram_client.search_findings(target_name, target_version)
        if findings:
            print(findings.summary_txt)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from ansible.template import Templar


@dataclass
class VariableManager(object):
    root_dir: str = ""

    def resolve(self, variables, data):
        templar = Templar(variables=variables)
        resolved_data = templar.template(data)
        return resolved_data



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass
from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    DefaultRiskType as RiskType,
    AnnotationCondition,
    Rule,
    Severity,
    RuleTag as Tag,
    RuleResult,
)


allow_url_list = ["https://*"]

deny_url_list = ["http://*"]


@dataclass
class InvalidDownloadSourceRule(Rule):
    rule_id: str = "R104"
    description: str = "A network transfer from unauthorized source is found."
    enabled: bool = True
    name: str = "InvalidDownloadSource"
    version: str = "v0.0.1"
    severity: Severity = Severity.HIGH
    tags: tuple = Tag.NETWORK

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        ac = AnnotationCondition().risk_type(RiskType.INBOUND)

        verdict = False
        detail = {}

        anno = task.get_annotation_by_condition(ac)
        if anno:
            if not self.is_allowed_url(anno.src.value, allow_url_list, deny_url_list):
                verdict = True
                detail["invalid_src"] = anno.src.value

        return RuleResult(verdict=verdict, detail=detail, file=task.file_info(), rule=self.get_metadata())

    def is_allowed_url(self, src, allow_list, deny_list):
        matched = True
        if len(allow_list) > 0:
            matched = False
            for a in allow_list:
                res = re.match(a, src)
                if res:
                    matched = True
        elif len(deny_list) > 0:
            for d in deny_list:
                res = re.match(d, src)
                if res:
                    matched = False
        return matched



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging


_logger = None

log_level_map = {
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


def set_logger_channel(channel: str = ""):
    global _logger
    _logger = logging.getLogger(channel)
    handler = logging.StreamHandler(sys.stdout)
    # default formatter
    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
    handler.setFormatter(formatter)
    _logger.addHandler(handler)


def set_log_level(level_str: str = "info"):
    global _logger
    level = log_level_map.get(level_str.lower(), None)
    _logger.setLevel(level)


def exception(*args, **kwargs):
    _logger.exception(*args, **kwargs)


def error(*args, **kwargs):
    _logger.error(*args, **kwargs)


def warning(*args, **kwargs):
    _logger.warning(*args, **kwargs)


def info(*args, **kwargs):
    _logger.info(*args, **kwargs)


def debug(*args, **kwargs):
    _logger.debug(*args, **kwargs)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from ansible_risk_insight.models import Annotation, RiskAnnotation, TaskCall, DefaultRiskType, FileChangeDetail
from ansible_risk_insight.annotators.module_annotator_base import ModuleAnnotator, ModuleAnnotatorResult


class LineInFileAnnotator(ModuleAnnotator):
    fqcn: str = "ansible.builtin.lineinfile"
    enabled: bool = True

    def run(self, task: TaskCall) -> List[Annotation]:
        path = task.args.get("path")
        mode = task.args.get("mode")
        unsafe_writes = task.args.get("unsafe_writes")
        state = task.args.get("state")

        annotation = RiskAnnotation.init(risk_type=DefaultRiskType.FILE_CHANGE,
                                         detail=FileChangeDetail(_path_arg=path, _state_arg=state, _mode_arg=mode, _unsafe_write_arg=unsafe_writes))
        return ModuleAnnotatorResult(annotations=[annotation])



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from ansible_risk_insight.models import Annotation, RiskAnnotation, TaskCall, DefaultRiskType, KeyConfigChangeDetail
from ansible_risk_insight.annotators.module_annotator_base import ModuleAnnotator, ModuleAnnotatorResult


class AptKeyAnnotator(ModuleAnnotator):
    fqcn: str = "ansible.builtin.apt_key"
    enabled: bool = True

    def run(self, task: TaskCall) -> List[Annotation]:
        # id = task.args.get("id")
        key = None
        if key is None:
            key = task.args.get("url")
        if key is None:
            key = task.args.get("data")
        if key is None:
            key = task.args.get("keyserver")

        state = task.args.get("state")

        annotation = RiskAnnotation.init(risk_type=DefaultRiskType.CONFIG_CHANGE,
                                         detail=KeyConfigChangeDetail(_key_arg=key, _state_arg=state))
        return ModuleAnnotatorResult(annotations=[annotation])



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List
from ansible_risk_insight.models import TaskCall, AnsibleRunContext, Annotation


class Annotator(object):
    type: str = ""
    context: AnsibleRunContext = None

    def __init__(self, context: AnsibleRunContext = None):
        if context:
            self.context = context

    def run(self, task: TaskCall):
        raise ValueError("this is a base class method")


@dataclass
class AnnotatorResult(object):
    annotations: List[Annotation] = None
    data: any = None

    def print(self):
        raise ValueError("this is a base class method")

    def to_json(self):
        raise ValueError("this is a base class method")

    def error(self):
        raise ValueError("this is a base class method")



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible_risk_insight.models import RiskAnnotation, TaskCall, DefaultRiskType, OutboundTransferDetail
from ansible_risk_insight.annotators.module_annotator_base import ModuleAnnotator, ModuleAnnotatorResult


class URIAnnotator(ModuleAnnotator):
    fqcn: str = "ansible.builtin.uri"
    enabled: bool = True

    def run(self, task: TaskCall) -> ModuleAnnotatorResult:
        method = task.args.get("method")

        annotations = []
        if method in ["PUT", "POST", "PATCH"]:
            url = task.args.get("url")
            body = task.args.get("body")
            annotation = RiskAnnotation.init(
                risk_type=DefaultRiskType.OUTBOUND,
                detail=OutboundTransferDetail(_dest_arg=url, _src_arg=body),
            )
            annotations.append(annotation)
        return ModuleAnnotatorResult(annotations=annotations)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from ansible_risk_insight.scanner import ARIScanner, config
from ansible_risk_insight.rules.R103_download_exec import DownloadExecRule


@pytest.mark.parametrize("type, name", [("project", "test/testdata/projects/my.collection")])
def test_scanner_with_project(type, name):
    s = _scan(type, name)
    risk_found_role_count = s.findings.report.get("summary", {}).get("roles", {}).get("risk_found", -1)
    assert risk_found_role_count > 0
    details = s.findings.report.get("details", [])
    assert len(details) > 0
    results = details[0].get("results", [])
    assert len(results) > 0
    download_exec_result = [r for r in results if r.get("rule", {}).get("name", "") == DownloadExecRule.name]
    assert len(download_exec_result) > 0


@pytest.mark.parametrize("type, name", [("collection", "community.mongodb")])
def test_scanner_with_collection(type, name):
    s = _scan(type, name)
    dep_names = [dep.get("name", "") for dep in s.findings.dependencies]
    assert len(dep_names) == 2
    assert "community.general" in dep_names
    assert "ansible.posix" in dep_names


@pytest.mark.parametrize("type, name", [("role", "test/testdata/roles/test_role")])
def test_scanner_with_role(type, name):
    s = _scan(type, name)
    risk_found_role_count = s.findings.report.get("summary", {}).get("roles", {}).get("risk_found", -1)
    assert risk_found_role_count > 0
    details = s.findings.report.get("details", [])
    assert len(details) > 0
    results = details[0].get("results", [])
    assert len(results) > 0
    download_exec_result = [r for r in results if r.get("rule", {}).get("name", "") == DownloadExecRule.name]
    assert len(download_exec_result) > 0


def _scan(type, name):
    s = ARIScanner(
        root_dir=config.data_dir,
        use_ansible_doc=False,
    )
    s.evaluate(
        type=type,
        name=name,
    )
    return s.get_last_scandata()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from ..scanner import ARIScanner, config
from ..utils import (
    is_url,
    is_local_path,
    get_collection_metadata,
    get_role_metadata,
    split_name_and_version,
)


class ARICLI:
    args = None

    def __init__(self):
        parser = argparse.ArgumentParser(description="TODO")
        parser.add_argument(
            "-s",
            "--save",
            action="store_true",
            help="enable file save under ARI_DATA_DIR (default=/tmp/ari-data)",
        )
        parser.add_argument("target_type", help="Content type", choices={"project", "role", "collection", "playbook", "taskfile"})
        parser.add_argument("target_name", help="Name")
        parser.add_argument("--playbook-only", action="store_true", help="if true, don't load playbooks/roles arround the specified playbook")
        parser.add_argument("--taskfile-only", action="store_true", help="if true, don't load playbooks/roles arround the specified taskfile")
        parser.add_argument("--skip-install", action="store_true", help="if true, skip install for the specified target")
        parser.add_argument("--dependency-dir", nargs="?", help="path to a directory that have dependencies for the target")
        parser.add_argument("--collection-name", nargs="?", help="if provided, use it as a collection name")
        parser.add_argument("--role-name", nargs="?", help="if provided, use it as a role name")
        parser.add_argument("--source", help="source server name in ansible config file (if empty, use public ansible galaxy)")
        parser.add_argument("--without-ram", action="store_true", help="if true, RAM data is not used and not even updated")
        parser.add_argument("--update-ram", action="store_true", help="if true, RAM data is not used for scan but updated with the scan result")
        parser.add_argument("--show-all", action="store_true", help="if true, show findings even if missing dependencies are found")
        parser.add_argument("--json", help="if specified, show findings in json format")
        parser.add_argument("--yaml", help="if specified, show findings in yaml format")
        parser.add_argument("-o", "--out-dir", help="output directory for the rule evaluation result")
        parser.add_argument(
            "-r", "--rules-dir", help=f"specify custom rule directories. use `-R` instead to ignore default rules in {config.rules_dir}"
        )
        parser.add_argument("-R", "--rules-dir-without-default", help="specify custom rule directories and ignore default rules")
        args = parser.parse_args()
        self.args = args

    def run(self):
        args = self.args
        target_name = args.target_name
        target_version = ""
        if args.target_type in ["collection", "role"]:
            target_name, target_version = split_name_and_version(target_name)

        collection_name = ""
        role_name = ""
        if args.collection_name:
            collection_name = args.collection_name

        if args.role_name:
            role_name = args.role_name

        is_local = False
        if args.target_type in ["collection", "role"] and is_local_path(target_name):
            is_local = True
        if args.target_type in ["project", "playbook", "taskfile"] and not is_url(target_name):
            is_local = True

        if is_local and not collection_name and not role_name:
            coll_meta = get_collection_metadata(target_name)
            if coll_meta:
                _namespace = coll_meta.get("collection_info", {}).get("namespace", "")
                _name = coll_meta.get("collection_info", {}).get("name", "")
                collection_name = f"{_namespace}.{_name}"

            role_meta = get_role_metadata(target_name)
            if role_meta:
                role_name = role_meta.get("galaxy_info", {}).get("role_name", "")

        rules_dir = config.rules_dir
        if args.rules_dir_without_default:
            rules_dir = args.rules_dir_without_default
        elif args.rules_dir:
            rules_dir = args.rules_dir + ":" + config.rules_dir

        silent = False
        pretty = False
        output_format = ""
        if args.json or args.yaml:
            silent = True
            pretty = True
            if args.json:
                output_format = "json"
            elif args.yaml:
                output_format = "yaml"

        read_ram = True
        write_ram = True
        if args.without_ram:
            read_ram = False
            write_ram = False
        elif args.update_ram:
            read_ram = False
            write_ram = True

        c = ARIScanner(
            root_dir=config.data_dir,
            rules_dir=rules_dir,
            do_save=args.save,
            read_ram=read_ram,
            write_ram=write_ram,
            show_all=args.show_all,
            silent=silent,
            pretty=pretty,
            output_format=output_format,
        )
        if not silent and not pretty:
            print("Start preparing dependencies")
        root_install = not args.skip_install
        if not silent and not pretty:
            print("Start scanning")
        c.evaluate(
            type=args.target_type,
            name=target_name,
            version=target_version,
            install_dependencies=root_install,
            dependency_dir=args.dependency_dir,
            collection_name=collection_name,
            role_name=role_name,
            source_repository=args.source,
            playbook_only=args.playbook_only,
            taskfile_only=args.taskfile_only,
            out_dir=args.out_dir,
        )



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from ansible_risk_insight.models import (
    AnsibleRunContext,
    RunTargetType,
    DefaultRiskType as RiskType,
    AnnotationCondition,
    Rule,
    Severity,
    RuleTag as Tag,
    RuleResult,
)


@dataclass
class DownloadExecRule(Rule):
    rule_id: str = "R103"
    description: str = "A downloaded file from parameterized source is executed"
    enabled: bool = True
    name: str = "Download & Exec"
    version: str = "v0.0.1"
    severity: Severity = Severity.HIGH
    tags: tuple = (Tag.NETWORK, Tag.COMMAND)

    def match(self, ctx: AnsibleRunContext) -> bool:
        return ctx.current.type == RunTargetType.Task

    def process(self, ctx: AnsibleRunContext):
        task = ctx.current

        verdict = False
        detail = {}

        ac = AnnotationCondition().risk_type(RiskType.CMD_EXEC)
        if task.has_annotation_by_condition(ac):
            cmd_an = task.get_annotation_by_condition(ac)
            if cmd_an:
                detail["command"] = cmd_an.command.raw

            ac2 = AnnotationCondition().risk_type(RiskType.INBOUND).attr("is_mutable_src", True)
            inbound_tasks = ctx.before(task).search(ac2)
            for inbound_task in inbound_tasks:
                inbound_an = inbound_task.get_annotation_by_condition(ac2)
                if not inbound_an:
                    continue
                detail["src"] = inbound_an.src.value

                # check if any of the exec_files are inside the download location
                # if so, the downloaded file is executed, so we report it
                download_location = inbound_an.dest
                executed_files = cmd_an.exec_files
                matched_files = [f for f in executed_files if f.is_inside(download_location)]
                if matched_files:
                    detail["executed_file"] = [f.value for f in matched_files]
                    verdict = True

        return RuleResult(verdict=verdict, detail=detail, file=task.file_info(), rule=self.get_metadata())



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy
import ansible_risk_insight.logger as logger
from .models import (
    Collection,
    Load,
    LoadType,
    Module,
    Play,
    Playbook,
    Repository,
    Role,
    Task,
    TaskFile,
    PlaybookFormatError,
    TaskFormatError,
)
from .model_loader import (
    load_collection,
    load_module,
    load_playbook,
    load_repository,
    load_role,
    load_taskfile,
)
from .utils import (
    split_target_playbook_fullpath,
    split_target_taskfile_fullpath,
    get_module_specs_by_ansible_doc,
)


class Parser:
    def __init__(self, do_save=False, use_ansible_doc=True, skip_playbook_format_error=True, skip_task_format_error=True):
        self.do_save = do_save
        self.use_ansible_doc = use_ansible_doc
        self.skip_playbook_format_error = skip_playbook_format_error
        self.skip_task_format_error = skip_task_format_error

    def run(self, load_data=None, load_json_path="", collection_name_of_project=""):
        ld = Load()
        if load_data is not None:
            ld = load_data
        elif load_json_path != "":
            if not os.path.exists(load_json_path):
                raise ValueError("file not found: {}".format(load_json_path))
            ld = Load.from_json(open(load_json_path, "r").read())

        collection_name = ""
        role_name = ""
        obj = None
        if ld.target_type == LoadType.COLLECTION:
            collection_name = ld.target_name
            try:
                obj = load_collection(
                    collection_dir=ld.path,
                    basedir=ld.path,
                    use_ansible_doc=self.use_ansible_doc,
                    skip_playbook_format_error=self.skip_playbook_format_error,
                    skip_task_format_error=self.skip_task_format_error,
                    load_children=False,
                )
            except PlaybookFormatError:
                if not self.skip_playbook_format_error:
                    raise
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception:
                logger.exception("failed to load the collection {}".format(collection_name))
                return
        elif ld.target_type == LoadType.ROLE:
            role_name = ld.target_name
            try:
                obj = load_role(
                    path=ld.path,
                    basedir=ld.path,
                    use_ansible_doc=self.use_ansible_doc,
                    skip_playbook_format_error=self.skip_playbook_format_error,
                    skip_task_format_error=self.skip_task_format_error,
                    load_children=False,
                )
            except PlaybookFormatError:
                if not self.skip_playbook_format_error:
                    raise
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception:
                logger.exception("failed to load the role {}".format(role_name))
                return
        elif ld.target_type == LoadType.PROJECT:
            repo_name = ld.target_name
            try:
                obj = load_repository(
                    path=ld.path,
                    basedir=ld.path,
                    use_ansible_doc=self.use_ansible_doc,
                    skip_playbook_format_error=self.skip_playbook_format_error,
                    skip_task_format_error=self.skip_task_format_error,
                )
            except PlaybookFormatError:
                if not self.skip_playbook_format_error:
                    raise
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception:
                logger.exception("failed to load the project {}".format(repo_name))
                return
            if obj.my_collection_name:
                collection_name = obj.my_collection_name
            if collection_name == "" and collection_name_of_project != "":
                collection_name = collection_name_of_project
        elif ld.target_type == LoadType.PLAYBOOK:
            basedir = ""
            target_playbook_path = ""
            if ld.playbook_yaml:
                target_playbook_path = ld.path
            else:
                basedir, target_playbook_path = split_target_playbook_fullpath(ld.path)
            playbook_name = ld.target_name
            try:
                if ld.playbook_only:
                    obj = load_playbook(
                        path=target_playbook_path,
                        yaml_str=ld.playbook_yaml,
                        basedir=basedir,
                        skip_playbook_format_error=self.skip_playbook_format_error,
                        skip_task_format_error=self.skip_task_format_error,
                    )
                else:
                    obj = load_repository(
                        path=basedir,
                        basedir=basedir,
                        target_playbook_path=target_playbook_path,
                        use_ansible_doc=self.use_ansible_doc,
                        skip_playbook_format_error=self.skip_playbook_format_error,
                        skip_task_format_error=self.skip_task_format_error,
                    )
            except PlaybookFormatError:
                if not self.skip_playbook_format_error:
                    raise
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception:
                logger.exception("failed to load the playbook {}".format(playbook_name))
                return
        elif ld.target_type == LoadType.TASKFILE:
            basedir = ""
            target_taskfile_path = ""
            if ld.taskfile_yaml:
                target_playbook_path = ld.path
            else:
                basedir, target_taskfile_path = split_target_taskfile_fullpath(ld.path)
            taskfile_name = ld.target_name
            try:
                if ld.taskfile_only:
                    obj = load_taskfile(
                        path=target_taskfile_path,
                        yaml_str=ld.taskfile_yaml,
                        basedir=basedir,
                        skip_task_format_error=self.skip_task_format_error,
                    )
                else:
                    obj = load_repository(
                        path=basedir,
                        basedir=basedir,
                        target_taskfile_path=target_taskfile_path,
                        use_ansible_doc=self.use_ansible_doc,
                        skip_playbook_format_error=self.skip_playbook_format_error,
                        skip_task_format_error=self.skip_task_format_error,
                    )
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception:
                logger.exception("failed to load the taskfile {}".format(taskfile_name))
                return
        else:
            raise ValueError("unsupported type: {}".format(ld.target_type))

        mappings = {
            "roles": [],
            "taskfiles": [],
            "modules": [],
            "playbooks": [],
        }

        basedir = ld.path
        if ld.target_type == LoadType.PLAYBOOK:
            basedir, _ = split_target_playbook_fullpath(ld.path)
        elif ld.target_type == LoadType.TASKFILE:
            basedir, _ = split_target_taskfile_fullpath(ld.path)

        roles = []
        for role_path in ld.roles:
            try:
                r = load_role(
                    path=role_path,
                    collection_name=collection_name,
                    basedir=basedir,
                    use_ansible_doc=self.use_ansible_doc,
                    skip_playbook_format_error=self.skip_playbook_format_error,
                    skip_task_format_error=self.skip_task_format_error,
                )
                roles.append(r)
            except PlaybookFormatError:
                if not self.skip_playbook_format_error:
                    raise
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception as e:
                logger.debug(f"failed to load a role: {e}")
                continue
            mappings["roles"].append([role_path, r.key])

        taskfiles = [tf for r in roles for tf in r.taskfiles]
        for taskfile_path in ld.taskfiles:
            try:
                tf = load_taskfile(
                    path=taskfile_path,
                    yaml_str=ld.taskfile_yaml,
                    role_name=role_name,
                    collection_name=collection_name,
                    basedir=basedir,
                    skip_task_format_error=self.skip_task_format_error,
                )
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception as e:
                logger.debug(f"failed to load a taskfile: {e}")
                continue
            taskfiles.append(tf)
            mappings["taskfiles"].append([taskfile_path, tf.key])

        playbooks = [p for r in roles for p in r.playbooks]
        for playbook_path in ld.playbooks:
            p = None
            try:
                p = load_playbook(
                    path=playbook_path,
                    yaml_str=ld.playbook_yaml,
                    role_name=role_name,
                    collection_name=collection_name,
                    basedir=basedir,
                    skip_playbook_format_error=self.skip_playbook_format_error,
                    skip_task_format_error=self.skip_task_format_error,
                )
            except PlaybookFormatError:
                if not self.skip_playbook_format_error:
                    raise
            except TaskFormatError:
                if not self.skip_task_format_error:
                    raise
            except Exception as e:
                logger.debug(f"failed to load a playbook: {e}")
                continue
            playbooks.append(p)
            mappings["playbooks"].append([playbook_path, p.key])

        plays = [play for p in playbooks for play in p.plays]

        tasks = [t for tf in taskfiles for t in tf.tasks]
        pre_tasks_in_plays = [t for p in plays for t in p.pre_tasks]
        tasks_in_plays = [t for p in plays for t in p.tasks]
        post_tasks_in_plays = [t for p in plays for t in p.post_tasks]
        tasks.extend(pre_tasks_in_plays)
        tasks.extend(tasks_in_plays)
        tasks.extend(post_tasks_in_plays)

        modules = [m for r in roles for m in r.modules]
        module_specs = {}
        if self.use_ansible_doc:
            module_specs = get_module_specs_by_ansible_doc(
                module_files=[fpath for fpath in ld.modules],
                fqcn_prefix=collection_name,
                search_path=ld.path,
            )

        for module_path in ld.modules:
            m = None
            try:
                m = load_module(
                    module_file_path=module_path,
                    role_name=role_name,
                    collection_name=collection_name,
                    basedir=basedir,
                    use_ansible_doc=self.use_ansible_doc,
                    module_specs=module_specs,
                )
            except Exception as e:
                logger.debug(f"failed to load a module: {e}")
                continue
            modules.append(m)
            mappings["modules"].append([module_path, m.key])

        logger.debug("roles: {}".format(len(roles)))
        logger.debug("taskfiles: {}".format(len(taskfiles)))
        logger.debug("modules: {}".format(len(modules)))
        logger.debug("playbooks: {}".format(len(playbooks)))
        logger.debug("plays: {}".format(len(plays)))
        logger.debug("tasks: {}".format(len(tasks)))

        collections = []
        projects = []
        if ld.target_type == LoadType.COLLECTION:
            collections = [obj]
        elif ld.target_type == LoadType.ROLE:
            pass
        elif ld.target_type == LoadType.PLAYBOOK:
            pass
        elif ld.target_type == LoadType.TASKFILE:
            pass
        elif ld.target_type == LoadType.PROJECT:
            projects = [obj]

        if len(collections) > 0:
            collections = [c.children_to_key() for c in collections]
        if len(projects) > 0:
            projects = [p.children_to_key() for p in projects]
        if len(roles) > 0:
            roles = [r.children_to_key() for r in roles]
        if len(taskfiles) > 0:
            taskfiles = [tf.children_to_key() for tf in taskfiles]
        if len(modules) > 0:
            modules = [m.children_to_key() for m in modules]
        if len(playbooks) > 0:
            playbooks = [p.children_to_key() for p in playbooks]
        if len(plays) > 0:
            plays = [p.children_to_key() for p in plays]
        if len(tasks) > 0:
            tasks = [t.children_to_key() for t in tasks]

        # save mappings
        ld.roles = mappings["roles"]
        ld.taskfiles = mappings["taskfiles"]
        ld.playbooks = mappings["playbooks"]
        ld.modules = mappings["modules"]

        definitions = {
            "collections": collections,
            "projects": projects,
            "roles": roles,
            "taskfiles": taskfiles,
            "modules": modules,
            "playbooks": playbooks,
            "plays": plays,
            "tasks": tasks,
        }

        return definitions, ld

    @classmethod
    def restore_definition_objects(cls, input_dir):

        collections = _load_object_list(Collection, os.path.join(input_dir, "collections.json"))

        # TODO: only repository?
        projects = _load_object_list(Repository, os.path.join(input_dir, "projects.json"))

        roles = _load_object_list(Role, os.path.join(input_dir, "roles.json"))

        taskfiles = _load_object_list(TaskFile, os.path.join(input_dir, "taskfiles.json"))

        modules = _load_object_list(Module, os.path.join(input_dir, "modules.json"))

        playbooks = _load_object_list(Playbook, os.path.join(input_dir, "playbooks.json"))

        plays = _load_object_list(Play, os.path.join(input_dir, "plays.json"))

        tasks = _load_object_list(Task, os.path.join(input_dir, "tasks.json"))

        definitions = {
            "collections": collections,
            "projects": projects,
            "roles": roles,
            "taskfiles": taskfiles,
            "modules": modules,
            "playbooks": playbooks,
            "plays": plays,
            "tasks": tasks,
        }

        ld = Load()
        mapping_path = os.path.join(input_dir, "mappings.json")
        if not os.path.exists(mapping_path):
            raise ValueError("file not found: {}".format(mapping_path))
        ld = Load.from_json(open(mapping_path, "r").read())
        return definitions, ld

    @classmethod
    def dump_definition_objects(cls, output_dir, definitions, ld):

        collections = definitions.get("collections", [])
        if len(collections) > 0:
            _dump_object_list(collections, os.path.join(output_dir, "collections.json"))
        projects = definitions.get("projects", [])
        if len(projects) > 0:
            _dump_object_list(projects, os.path.join(output_dir, "projects.json"))

        roles = definitions.get("roles", [])
        if len(roles) > 0:
            _dump_object_list(roles, os.path.join(output_dir, "roles.json"))

        taskfiles = definitions.get("taskfiles", [])
        if len(taskfiles) > 0:
            _dump_object_list(taskfiles, os.path.join(output_dir, "taskfiles.json"))

        modules = definitions.get("modules", [])
        if len(modules) > 0:
            _dump_object_list(modules, os.path.join(output_dir, "modules.json"))

        playbooks = definitions.get("playbooks", [])
        if len(playbooks) > 0:
            _dump_object_list(playbooks, os.path.join(output_dir, "playbooks.json"))

        plays = definitions.get("plays", [])
        if len(plays) > 0:
            _dump_object_list(plays, os.path.join(output_dir, "plays.json"))

        tasks = definitions.get("tasks", [])
        if len(tasks) > 0:
            _dump_object_list(tasks, os.path.join(output_dir, "tasks.json"))

        mapping_path = os.path.join(output_dir, "mappings.json")
        open(mapping_path, "w").write(ld.dump())


def _dump_object_list(obj_list, output_path):
    tmp_obj_list = copy.deepcopy(obj_list)
    lines = []
    for i in range(len(tmp_obj_list)):
        lines.append(tmp_obj_list[i].dump())
    open(output_path, "w").write("\n".join(lines))
    return


def _load_object_list(cls, input_path):
    obj_list = []
    if os.path.exists(input_path):
        with open(input_path, "r") as f:
            for line in f:
                obj = cls.from_json(line)
                obj_list.append(obj)
    return obj_list


def load_name2target_name(path):
    filename = os.path.basename(path)
    parts = os.path.splitext(filename)
    prefix = "load-"
    target_name = parts[0]
    if target_name.startswith(prefix):
        target_name = target_name[len(prefix) :]
    return target_name


# if __name__ == "__main__":
#     parser = argparse.ArgumentParser(
#         prog="parser.py",
#         description=(
#             "parse collection/role and its children and output definition"
#             " json"
#         ),
#         epilog="end",
#         add_help=True,
#     )

#     parser.add_argument(
#         "-l", "--load-path", default="", help="load json path/dir"
#     )
#     parser.add_argument(
#         "-i",
#         "--index-path",
#         default="",
#         help=(
#             "if specified, load files in this index.json (--load-path will be"
#             " ignored)"
#         ),
#     )
#     parser.add_argument(
#         "--root",
#         action="store_true",
#         help="enable this if the target is the root",
#     )
#     parser.add_argument(
#         "--ext",
#         action="store_true",
#         help="enable this if the target is the external dependency(s)",
#     )
#     parser.add_argument(
#         "-o", "--output-dir", default="", help="path to the output dir"
#     )

#     args = parser.parse_args()

#     if not args.root and not args.ext:
#         logger.error('either "--root" or "--ext" must be specified')
#         sys.exit(1)
#     is_ext = args.ext

#     if args.load_path == "" and args.index_path == "":
#         logger.error("either `--load-path` or `--index-path` is required")
#         sys.exit(1)

#     if args.root and args.load_path == "":
#         logger.error('"--load-path" must be specified for "--root" mode')
#         sys.exit(1)

#     if args.root and not os.path.isfile(args.load_path):
#         logger.error(
#             '"--load-path" must be a single .json file for "--root" mode'
#         )
#         sys.exit(1)

#     if args.load_path != "" and not os.path.exists(args.load_path):
#         logger.error("No such file or directory: {}".format(args.load_path))
#         sys.exit(1)

#     if args.index_path != "" and not os.path.exists(args.index_path):
#         logger.error("No such file or directory: {}".format(args.index_path))
#         sys.exit(1)

#     load_json_path_list = []
#     if args.index_path != "":
#         if os.path.isfile(args.index_path):
#             with open(args.index_path, "r") as file:
#                 index_data = json.load(file)
#                 load_dir = index_data.get("out_path", "")
#                 load_json_name_list = index_data.get(
#                     "generated_load_files", []
#                 )
#                 load_json_name_list = [
#                     f.get("file", "") if isinstance(f, dict) else f
#                     for f in load_json_name_list
#                 ]
#                 load_json_path_list = [
#                     os.path.join(load_dir, f) for f in load_json_name_list
#                 ]
#         else:
#             files = os.listdir(args.index_path)
#             index_json_path_list = [
#                 os.path.join(args.index_path, fname)
#                 for fname in files
#                 if fname.startswith("index-") and fname.endswith(".json")
#             ]
#             for i in index_json_path_list:
#                 with open(i, "r") as file:
#                     index_data = json.load(file)
#                     load_dir = index_data.get("out_path", "")
#                     load_json_name_list = index_data.get(
#                         "generated_load_files", []
#                     )
#                     load_json_name_list = [
#                         f.get("file", "") if isinstance(f, dict) else f
#                         for f in load_json_name_list
#                     ]
#                     tmp_load_json_list = [
#                         os.path.join(load_dir, f) for f in load_json_name_list
#                     ]
#                     for load_json_path in tmp_load_json_list:
#                         if load_json_path not in load_json_path_list:
#                             load_json_path_list.append(load_json_path)
#     elif args.load_path != "":
#         if os.path.isfile(args.load_path):
#             load_json_path_list = [args.load_path]
#         else:
#             files = os.listdir(args.load_path)
#             load_json_path_list = [
#                 os.path.join(args.load_path, fname)
#                 for fname in files
#                 if fname.startswith("load-") and fname.endswith(".json")
#             ]

#     if len(load_json_path_list) == 0:
#         logger.info("no load json files found. exitting.")
#         sys.exit()

#     profiles = [
#         (
#             load_json_path,
#             os.path.join(
#                 args.output_dir, load_name2target_name(load_json_path)
#             )
#             if is_ext
#             else args.output_dir,
#         )
#         for load_json_path in load_json_path_list
#     ]

#     num = len(profiles)
#     if num == 0:
#         logger.info("no load json files found. exitting.")
#         sys.exit()
#     else:
#         logger.info("start parsing {} target(s)".format(num))
#     p = Parser()

#     def parse_single(single_input):
#         i = single_input[0]
#         num = single_input[1]
#         load_json_path = single_input[2]
#         output_dir = single_input[3]
#         print("[{}/{}] {}       ".format(i + 1, num, load_json_path))

#         definitions, ld = p.run(load_json_path=load_json_path)

#         if output_dir == "":
#             raise ValueError("Invalid output_dir")
#         if os.path.exists(output_dir):
#             os.makedirs(output_dir, exist_ok=True)
#         Parser.dump_definition_objects(output_dir, definitions, ld)


#     parallel_input_list = [
#         (i, num, load_json_path, output_dir)
#         for i, (load_json_path, output_dir) in enumerate(profiles)
#     ]
#     _ = joblib.Parallel(n_jobs=-1)(
#         joblib.delayed(parse_single)(single_input)
#         for single_input in parallel_input_list
#     )



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import re


valid_playbook_re = re.compile(r"^\s*?-?\s*?(?:hosts|include|import_playbook):\s*?.*?$")


# this method is based on awx code
# awx/main/utils/ansible.py#L42-L64 in ansible/awx
def could_be_playbook(fpath):
    basename, ext = os.path.splitext(fpath)
    if ext not in [".yml", ".yaml"]:
        return False
    # Filter files that do not have either hosts or top-level
    # includes. Use regex to allow files with invalid YAML to
    # show up.
    matched = False
    try:
        with codecs.open(fpath, "r", encoding="utf-8", errors="ignore") as f:
            for n, line in enumerate(f):
                if valid_playbook_re.match(line):
                    matched = True
                    break
                # Any YAML file can also be encrypted with vault;
                # allow these to be used as the main playbook.
                elif n == 0 and line.startswith("$ANSIBLE_VAULT;"):
                    matched = True
                    break
    except IOError:
        return False
    return matched


# this method is based on awx code
# awx/main/models/projects.py#L206-L217 in ansible/awx
def search_playbooks(root_path):
    results = []
    if root_path and os.path.exists(root_path):
        for dirpath, dirnames, filenames in os.walk(root_path, followlinks=False):
            if skip_directory(dirpath):
                continue
            for filename in filenames:
                fpath = os.path.join(dirpath, filename)
                if could_be_playbook(fpath):
                    results.append(fpath)
    return sorted(results, key=lambda x: x.lower())


# this method is based on awx code
# awx/main/utils/ansible.py#L24-L39 in ansible/awx
def skip_directory(relative_directory_path):
    path_elements = relative_directory_path.split(os.sep)
    # Exclude files in a roles subdirectory.
    if "roles" in path_elements:
        return True
    # Filter files in a tasks subdirectory.
    if "tasks" in path_elements:
        return True
    for element in path_elements:
        # Do not include dot files or dirs
        if element.startswith("."):
            return True
    # Exclude anything inside of group or host vars directories
    if "group_vars" in path_elements or "host_vars" in path_elements:
        return True
    return False



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible_risk_insight.models import TaskCall
from ansible_risk_insight.annotators.risk_annotator_base import RiskAnnotator


class AnsibleBuiltinRiskAnnotator(RiskAnnotator):
    name: str = "ansible.builtin"
    enabled: bool = True

    def match(self, task: TaskCall) -> bool:
        resolved_name = task.spec.resolved_name
        return resolved_name.startswith("ansible.builtin.")

    # embed "analyzed_data" field in Task
    def run(self, task: TaskCall):
        if not self.match(task):
            return []

        return self.run_module_annotators("ansible.builtin", task)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible_risk_insight.models import RiskAnnotation, TaskCall, DefaultRiskType, CommandExecDetail
from ansible_risk_insight.annotators.module_annotator_base import ModuleAnnotator, ModuleAnnotatorResult


class RawAnnotator(ModuleAnnotator):
    fqcn: str = "ansible.builtin.raw"
    enabled: bool = True

    def run(self, task: TaskCall) -> ModuleAnnotatorResult:
        cmd = task.args.get("")
        if cmd is None:
            cmd = task.args.get("command")
        if cmd is None:
            cmd = task.args.get("cmd")
        if cmd is None:
            cmd = task.args.get("argv")

        annotation = RiskAnnotation.init(
            risk_type=DefaultRiskType.CMD_EXEC,
            detail=CommandExecDetail(command=cmd),
        )
        return ModuleAnnotatorResult(annotations=[annotation])



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from ansible_risk_insight.models import Annotation, RiskAnnotation, TaskCall, DefaultRiskType, InboundTransferDetail
from ansible_risk_insight.annotators.module_annotator_base import ModuleAnnotator, ModuleAnnotatorResult
from ansible.module_utils.parsing.convert_bool import boolean


class UnarchiveAnnotator(ModuleAnnotator):
    fqcn: str = "ansible.builtin.unarchive"
    enabled: bool = True

    def run(self, task: TaskCall) -> List[Annotation]:
        src = task.args.get("src")  # required
        dest = task.args.get("dest")  # required
        remote_src = task.args.get("remote_src")

        is_remote_src = False
        if remote_src:
            is_remote_src = boolean(remote_src.raw) or boolean(remote_src.templated)

        url_sep = "://"
        is_download = False
        if is_remote_src and (url_sep in src.raw or url_sep in src.templated):
            is_download = True

        if not is_download:
            return None

        annotation = RiskAnnotation.init(risk_type=DefaultRiskType.INBOUND, detail=InboundTransferDetail(_src_arg=src, _dest_arg=dest))
        return ModuleAnnotatorResult(annotations=[annotation])



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from .search import RAMSearchCLI
from .list import RAMListCLI
from .diff import RAMDiffCLI
from .generate import RAMGenerateCLI
from .update import RAMUpdateCLI
from .release import RAMReleaseCLI


ram_actions = ["search", "list", "diff", "generate", "update", "release"]


class RAMCLI:
    _cli = None

    def __init__(self):

        args = sys.argv
        if len(args) > 2:
            action = args[2]
            # "search" can be abbreviated
            if action not in ram_actions:
                action = "search"
                target_name = sys.argv[2]
                sys.argv[2] = action
                sys.argv.insert(3, target_name)

            if action == "search":
                self._cli = RAMSearchCLI()
            elif action == "list":
                self._cli = RAMListCLI()
            elif action == "diff":
                self._cli = RAMDiffCLI()
            elif action == "generate":
                self._cli = RAMGenerateCLI()
            elif action == "update":
                self._cli = RAMUpdateCLI()
            elif action == "release":
                self._cli = RAMReleaseCLI()
            else:
                raise ValueError(f"The action {action} is not supported")
        else:
            raise ValueError(f"An action must be specified; {ram_actions}")

    def run(self):
        if self._cli:
            self._cli.run()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from .cli import ARICLI
from .cli.ram import RAMCLI
from ansible_risk_insight.scanner import ARIScanner, Config

ari_actions = ["project", "playbook", "collection", "role", "taskfile"]
ram_actions = ["ram"]

all_actions = ari_actions + ram_actions


def main():
    if len(sys.argv) == 1:

        print("Please specify one of the following operations of ari.")
        print("[operations]")
        print("   playbook     scan a playbook (e.g. `ari playbook path/to/playbook.yml` )")
        print("   collection   scan a collection (e.g. `ari collection collection.name` )")
        print("   role         scan a role (e.g. `ari role role.name` )")
        print("   project      scan a project (e.g. `ari project path/to/project`)")
        print("   taskfile     scan a taskfile (e.g. `ari taskfile path/to/taskfile.yml`)")
        print("   ram          operate the backend data (e.g. `ari ram generate -f input.txt`)")
        sys.exit()

    action = sys.argv[1]

    if action in ari_actions:
        cli = ARICLI()
        cli.run()
    elif action == "ram":
        cli = RAMCLI()
        cli.run()
    else:
        print(f"The action {action} is not supported!", file=sys.stderr)
        sys.exit(1)


__all__ = ["ARIScanner", "Config", "models"]



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from dataclasses import dataclass
from typing import List
from ansible_risk_insight.keyutil import detect_type
from ansible_risk_insight.models import (
    ObjectList,
    Repository,
    Playbook,
    Role,
    TaskCall,
    VariableAnnotation,
    Arguments,
    ArgumentsType,
    Variable,
    VariableType,
)
from ansible_risk_insight.context import Context, resolve_module_options
from ansible_risk_insight.annotators.annotator_base import Annotator, AnnotatorResult


class VariableAnnotator(Annotator):
    type: str = VariableAnnotation.type
    context: Context = None

    def __init__(self, context: Context):
        self.context = context

    def run(self, taskcall: TaskCall):
        resolved = resolve_module_options(self.context, taskcall)
        resolved_module_options = resolved[0]
        resolved_variables = resolved[1]
        # mutable_vars_per_mo = resolved[2]
        used_variables = resolved[3]
        _vars = []
        is_mutable = False
        for rv in resolved_variables:
            v_name = rv.get("key", "")
            v_value = rv.get("value", "")
            v_type = rv.get("type", VariableType.Unknown)
            elements = []
            if v_name in used_variables:
                if not isinstance(used_variables[v_name], dict):
                    continue
                for u_v_name, info in used_variables[v_name].items():
                    if u_v_name == v_name:
                        continue
                    u_v_value = info.get("value", "")
                    u_v_type = info.get("type", VariableType.Unknown)
                    u_v = Variable(
                        name=u_v_name,
                        value=u_v_value,
                        type=u_v_type,
                        used_in=taskcall.key,
                    )
                    elements.append(u_v)
            v = Variable(
                name=v_name,
                value=v_value,
                type=v_type,
                elements=elements,
                used_in=taskcall.key,
            )
            _vars.append(v)
            if v.is_mutable:
                is_mutable = True

        for v in _vars:
            history = self.context.var_use_history.get(v.name, [])
            history.append(v)
            self.context.var_use_history[v.name] = history

        m_opts = taskcall.spec.module_options
        if isinstance(m_opts, list):
            args_type = ArgumentsType.LIST
        elif isinstance(m_opts, dict):
            args_type = ArgumentsType.DICT
        else:
            args_type = ArgumentsType.SIMPLE
        args = Arguments(
            type=args_type,
            raw=m_opts,
            vars=_vars,
            resolved=True,  # TODO: False if not resolved
            templated=resolved_module_options,
            is_mutable=is_mutable,
        )
        taskcall.args = args
        # deep copy the history here because the context is updated by subsequent taskcalls
        taskcall.variable_set = deepcopy(self.context.var_set_history)
        taskcall.variable_use = deepcopy(self.context.var_use_history)
        taskcall.become = self.context.become

        return VariableAnnotatorResult()


@dataclass
class VariableAnnotatorResult(AnnotatorResult):
    pass


def tree_to_task_list(tree, node_objects):
    node_dict = {}
    for no in node_objects.items:
        node_dict[no.key] = no

    def getSubTree(node):
        tasks = []
        resolved_name = ""
        no = node_dict[node.key]
        node_type = detect_type(node.key)

        children_tasks = []
        if node_type == "module":
            resolved_name = no.fqcn
        elif node_type == "role":
            resolved_name = no.fqcn
        elif node_type == "taskfile":
            resolved_name = no.key

        children_per_type = {}
        for c in node.children:
            ctype = detect_type(c.key)
            if ctype in children_per_type:
                children_per_type[ctype].append(c)
            else:
                children_per_type[ctype] = [c]

        # obj["children_types"] = list(children_per_type.keys())
        if "playbook" in children_per_type:
            tasks_per_children = [getSubTree(c) for c in children_per_type["playbook"]]
            for (_tasks, _) in tasks_per_children:
                children_tasks.extend(_tasks)
        if "play" in children_per_type:
            tasks_per_children = [getSubTree(c) for c in children_per_type["play"]]
            for (_tasks, _) in tasks_per_children:
                children_tasks.extend(_tasks)
        if "role" in children_per_type:
            tasks_per_children = [getSubTree(c) for c in children_per_type["role"]]
            for (_tasks, _) in tasks_per_children:
                children_tasks.extend(_tasks)
            if node_type == "task":
                fqcns = [fqcn for (_, fqcn) in tasks_per_children]
                resolved_name = fqcns[0] if len(fqcns) > 0 else ""
        if "taskfile" in children_per_type:
            tasks_per_children = [getSubTree(c) for c in children_per_type["taskfile"]]
            for (_tasks, _) in tasks_per_children:
                children_tasks.extend(_tasks)
            if node_type == "task":
                _tf_path_list = [_tf_path for (_, _tf_path) in tasks_per_children]
                resolved_name = _tf_path_list[0] if len(_tf_path_list) > 0 else ""
        if "task" in children_per_type:
            tasks_per_children = [getSubTree(c) for c in children_per_type["task"]]
            for (_tasks, _) in tasks_per_children:
                children_tasks.extend(_tasks)
        if "module" in children_per_type:
            if node_type == "task":
                fqcns = [getSubTree(c)[1] for c in children_per_type["module"]]
                resolved_name = fqcns[0] if len(fqcns) > 0 else ""

        if node_type == "task":
            no.resolved_name = resolved_name
            tasks.append(no.__dict__)
        tasks.extend(children_tasks)
        return tasks, resolved_name

    tasks, _ = getSubTree(tree)
    return tasks


def resolve_variables(tree: ObjectList, additional: ObjectList) -> List[TaskCall]:
    tree_root_key = tree.items[0].spec.key if len(tree.items) > 0 else ""
    inventories = get_inventories(tree_root_key, additional)
    context = Context(inventories=inventories)
    depth_dict = {}
    resolved_taskcalls = []
    for call_obj in tree.items:
        caller_depth_lvl = 0
        if call_obj.called_from != "":
            caller_key = call_obj.called_from
            caller_depth_lvl = depth_dict.get(caller_key, 0)
        depth_lvl = caller_depth_lvl + 1
        depth_dict[call_obj.key] = depth_lvl
        context.add(call_obj, depth_lvl)
        if isinstance(call_obj, TaskCall):
            result = VariableAnnotator(context=context).run(call_obj)
            if not result:
                continue
            if result.annotations:
                call_obj.annotations.extend(result.annotations)
            resolved_taskcalls.append(call_obj)
    return resolved_taskcalls


def get_inventories(tree_root_key, additional):
    if tree_root_key == "":
        return []
    tree_root_type = detect_type(tree_root_key)
    projects = additional.find_by_type("repository")
    inventories = []
    found = False
    for p in projects:
        if not isinstance(p, Repository):
            continue
        if tree_root_type == "playbook":
            for playbook in p.playbooks:
                if isinstance(playbook, str):
                    if playbook == tree_root_key:
                        inventories = p.inventories
                        found = True
                elif isinstance(playbook, Playbook):
                    if playbook.key == tree_root_key:
                        inventories = p.inventories
                        found = True
                if found:
                    break
        elif tree_root_type == "role":
            for role in p.roles:
                if isinstance(role, str):
                    if role == tree_root_key:
                        inventories = p.inventories
                        found = True
                elif isinstance(role, Role):
                    if role.key == tree_root_key:
                        inventories = p.inventories
                        found = True
                if found:
                    break
        if found:
            break
    return inventories
